import importlib
import re
from enum import Enum

import ibm_watsonx_data_integration.services.datastage.models.flow_json_model as models
import ibm_watsonx_data_integration.services.datastage.models.schema.field as Field
from ibm_watsonx_data_integration.services.datastage.codegen.datasource_mappings import DATASOURCE_MAPPINGS
from ibm_watsonx_data_integration.services.datastage.codegen.label_mappings import LABEL_MAPPINGS
from ibm_watsonx_data_integration.services.datastage.codegen.op_mappings import OP_MAPPINGS
from ibm_watsonx_data_integration.services.datastage.models.flow.dag import (
    DAG,
    Link,
    Node,
    StageNode,
)
from ibm_watsonx_data_integration.services.datastage.models.flow.datastage_flow import DataStageFlow
from ibm_watsonx_data_integration.services.datastage.models.schema.field.base_field import BaseField
from ibm_watsonx_data_integration.services.datastage.models.schema.schema import Schema
from ibm_watsonx_data_integration.services.datastage.models.stage_models.complex_stages import lookup
from ibm_watsonx_data_integration.services.datastage.models.stage_models.complex_stages.complex_flat_file import (
    Column,
    OutputColumns,
    Record,
    RecordID,
    Constraint as CFFConstraint,
)

from ibm_watsonx_data_integration.services.datastage.models.stage_models.complex_stages.rest import (
    Authentication,
    Body,
    Control,
    Cookie,
    FormData,
    FormURLEncodedData,
    Header,
    Parameter,
    Request,
    RequestInfo,
    Response,
    Settings,
    Variable,
)
from ibm_watsonx_data_integration.services.datastage.models.stage_models.complex_stages.transformer import (
    Constraint,
    LoopVariable,
    StageVariable,
    Trigger,
)

CONNECTIONS = {}
JAVA_LIBRARIES = {}
FUNCTION_LIBRARIES = {}
MATCH_SPECIFICATIONS = {}


def label_to_stage_name(label: str) -> str:
    words = re.findall(r"[a-zA-Z0-9]+", label)
    result = []
    for word in words:
        if word.isupper():
            result.append(word)
        else:
            result.append(word.capitalize())
    return "".join(result)


def op_name_to_conn_class(op_name: str):
    """Converts an operator name (usually snake_case) into the PascalCase class name of that stage."""
    split = op_name.split("_")
    caps = [part[0].upper() + part[1:] for part in split]
    return "".join(caps) + "Conn"


def instantiate_stage_model(model_class: str, model_stage_props: dict):
    """Instantiates a model object of a specific model class using the given dictionary of properties."""
    mod = importlib.import_module(f"ibm_watsonx_data_integration.services.datastage.models.stage_models.{model_class}")
    class_ = getattr(mod, model_class)
    obj = class_.model_construct(**model_stage_props)
    # Convert enums from strings to objects
    for prop, value in obj.model_fields.items():
        if "enum" in str(value.annotation) or isinstance(value.default, Enum):
            try:
                enum = type(value.default)(getattr(obj, prop))
                setattr(obj, prop, enum)
            except (ValueError, KeyError, TypeError):
                pass
    return obj


def instantiate_stage_node(op_name: str, dag: DAG, model: models.BaseModel) -> StageNode:
    """Instantiates a stage object of a specific stage class."""
    mod = importlib.import_module(f"ibm_watsonx_data_integration.services.datastage.models.stages.{op_name}")
    label = LABEL_MAPPINGS[op_name]
    class_ = getattr(mod, label_to_stage_name(label) + "Stage")
    obj = class_(dag, model)
    return obj


# edit this to keep track of same connection instance
def instantiate_connection_model(op_name: str, conn_props: dict):
    if conn_props["name"] in CONNECTIONS:
        return CONNECTIONS[conn_props["name"]]
    mod = importlib.import_module(f"ibm_watsonx_data_integration.services.datastage.models.connections.{op_name}_connection")
    class_ = getattr(mod, op_name_to_conn_class(op_name))
    obj = class_.model_construct(**conn_props)
    CONNECTIONS[obj.name] = obj
    return obj


# def instantiate_java_library_model(jl_props: dict):
#     if jl_props["name"] in JAVA_LIBRARIES:
#         return JAVA_LIBRARIES[jl_props["name"]]
#     jl = JavaLibrary(**jl_props)
#     JAVA_LIBRARIES[jl.name] = jl
#     return jl


# def instantiate_function_library_model(fl_props: dict):
#     if fl_props["name"] in FUNCTION_LIBRARIES:
#         return FUNCTION_LIBRARIES[fl_props["name"]]
#     fl = FunctionLibrary(**fl_props)
#     FUNCTION_LIBRARIES[fl.name] = fl
#     return fl


# def instantiate_match_specification_model(ms_props: dict):
#     if ms_props["name"] in MATCH_SPECIFICATIONS:
#         return MATCH_SPECIFICATIONS[ms_props["name"]]
#     ms = MatchSpecification(**ms_props)
#     MATCH_SPECIFICATIONS[ms.name] = ms
#     return ms


class DAGGenerator:
    def __init__(self, flow_model: models.Flow):
        self.flow_model = flow_model
        self.id_to_dag_node: dict[str, Node] = {}
        self.id_to_node_model: dict[str, models.Node] = {}
        self.schemas: dict[str, Schema] = {}
        self.fc = DataStageFlow()
        self.dag = self.fc._dag
        self.rcp = {}

    def generate(self) -> DataStageFlow:
        for record_schema in self.flow_model.schemas:
            self.process_schema(record_schema)

        for pipeline_model in self.flow_model.pipelines:
            if self.flow_model.primary_pipeline == pipeline_model.id:
                self.process_pipeline(pipeline_model)

        # if self.flow_model.external_paramsets:
        #     for paramset in self.flow_model.external_paramsets:
        #         self.fc.use_paramset(self.process_paramsets(paramset))

        # if self.flow_model.parameters and "local_parameters" in self.flow_model.parameters:
        #     self.fc.use_localparams(self.process_localparams(self.flow_model.parameters["local_parameters"]))
        # if hasattr(self.flow_model.app_data, "datastage") and self.flow_model.app_data.datastage:
        #     if "message_handlers" in self.flow_model.app_data.datastage:
        #         self.fc.use_local_message_handler(
        #             self.process_local_message_handler(self.flow_model.app_data.datastage["message_handlers"])
        #         )
        #     if "external_message_handlers" in self.flow_model.app_data.datastage:
        #         for asset_id in self.flow_model.app_data.datastage["external_message_handlers"]:
        #             self.fc.use_message_handler(self.process_message_handler(asset_id))

        # self.fc.use_runtime(self.process_runtime(self.flow_model.app_data))
        self.fc._dag = self.dag

        return self.fc

    def process_schema(self, schema_model: models.RecordSchema) -> None:
        fields = [self.generate_field(field_model) for field_model in schema_model.fields]
        # if not len(fields):
        #     print(f"WARNING: Found schema with no fields, skipping: {schema_model.id}")
        #     return

        schema = Schema(fields)

        assert schema_model.id not in self.schemas
        self.schemas[schema_model.id] = schema

    def generate_field(self, field_model: models.FieldModel) -> BaseField:
        ODBC_TO_CLASS = {
            "BIGINT": Field.BigInt,
            "BINARY": Field.Binary,
            "BIT": Field.Bit,
            "CHAR": Field.Char,
            "DATE": Field.Date,
            "DECIMAL": Field.Decimal,
            "DOUBLE": Field.Double,
            "FLOAT": Field.Float,
            "INTEGER": Field.Integer,
            "LONGVARBINARY": Field.LongVarBinary,
            "LONGVARCHAR": Field.LongVarChar,
            "LONGNVARCHAR": Field.LongVarChar,
            "NUMERIC": Field.Numeric,
            "REAL": Field.Real,
            "SMALLINT": Field.SmallInt,
            "TIME": Field.Time,
            "TIMESTAMP": Field.Timestamp,
            "TINYINT": Field.TinyInt,
            "UNKNOWN": Field.Unknown,
            "VARBINARY": Field.VarBinary,
            "VARCHAR": Field.VarChar,
            "NCHAR": Field.NChar,
            "NVARCHAR": Field.VarChar,
            "WCHAR": Field.NChar,
            "WLONGVARCHAR": Field.LongNVarChar,
            "WVARCHAR": Field.NVarChar,
        }

        try:
            odbc = field_model.app_data["odbc_type"]
        except Exception:
            raise KeyError(f"Field {field_model.name} is missing an ODBC type")

        try:
            class_ = ODBC_TO_CLASS[odbc.upper()]
        except KeyError:
            raise KeyError(f"Unknown ODBC type for field {field_model.name}: {odbc}")

        field_obj: BaseField = class_(field_model.name)
        field_obj.configuration = models.FieldModelComplex.from_field_model(field_model)

        return field_obj

    def process_pipeline(self, pipeline_model: models.Pipeline) -> None:
        for node_model in pipeline_model.nodes:
            model_stage_node = self.generate_node(node_model)
            assert node_model.id
            self.id_to_dag_node[node_model.id] = model_stage_node
            self.id_to_node_model[node_model.id] = node_model

        for node_id in self.id_to_dag_node:
            self.link_up_node(self.id_to_dag_node[node_id], self.id_to_node_model[node_id])

        for link in self.dag.links():
            if link.name in self.rcp:
                try:
                    if hasattr(link.src, "configuration"):
                        link.src.configuration.runtime_column_propagation = self.rcp[link.name]
                    elif hasattr(link.src, "rcp"):
                        link.src.rcp = self.rcp[link.name]
                    else:
                        link.src.with_runtime_column_propagation(self.rcp[link.name])
                except Exception:
                    pass

        self.populate_funnel_schema(pipeline_model=pipeline_model)
        self.populate_difference_schema(pipeline_model=pipeline_model)
        self.populate_sort_schema(pipeline_model=pipeline_model)
        self.populate_pivot_schema(pipeline_model=pipeline_model)
        self.populate_change_capture_schema(pipeline_model=pipeline_model)
        self.populate_transformer_schema(pipeline_model=pipeline_model)
        self.populate_rest_schema()

        # if pipeline_model.app_data.ui_data and pipeline_model.app_data.ui_data.comments:
        #     self.process_comments(pipeline_model=pipeline_model)

    def populate_funnel_schema(self, pipeline_model: models.Pipeline) -> None:
        for link in self.dag.links():
            if hasattr(link.src, "configuration"):
                if link.src.configuration.op_name == "PxFunnel":
                    for node_model in pipeline_model.nodes:
                        if hasattr(node_model, "op") and node_model.op == "PxFunnel":
                            if node_model.app_data.ui_data.label == link.src.label:
                                for output in node_model.outputs:
                                    if hasattr(output, "parameters") and output.parameters and "valueDerivation" in output.parameters:
                                        for vd in output.parameters["valueDerivation"]:
                                            if vd["columnName"] != vd["parsedExpression"]:
                                                for field in link.schema.fields:
                                                    if field.configuration.name == vd["columnName"]:
                                                        field.configuration.metadata.source_field_id = vd["parsedExpression"]

    def populate_difference_schema(self, pipeline_model: models.Pipeline) -> None:
        for link in self.dag.links():
            if hasattr(link.src, "configuration"):
                if link.src.configuration.op_name == "PxDifference":
                    for node_model in pipeline_model.nodes:
                        if hasattr(node_model, "op") and node_model.op == "PxDifference":
                            if node_model.app_data.ui_data.label == link.src.label:
                                for output in node_model.outputs:
                                    if hasattr(output, "parameters") and output.parameters and "valueDerivation" in output.parameters:
                                        for vd in output.parameters["valueDerivation"]:
                                            if vd["parsedExpression"] == "DiffCode()":
                                                if vd["columnName"] != "diff":
                                                    for field in link.schema.fields:
                                                        if field.configuration.name == vd["columnName"]:
                                                            field.configuration.app_data["difference"] = True

    def populate_sort_schema(self, pipeline_model: models.Pipeline) -> None:
        for link in self.dag.links():
            if hasattr(link.src, "configuration"):
                if link.src.configuration.op_name == "PxSort":
                    for node_model in pipeline_model.nodes:
                        if hasattr(node_model, "op") and node_model.op == "PxSort":
                            if node_model.app_data.ui_data.label == link.src.label:
                                for output in node_model.outputs:
                                    if hasattr(output, "parameters") and output.parameters and "valueDerivation" in output.parameters:
                                        for vd in output.parameters["valueDerivation"]:
                                            if vd["parsedExpression"] == "ClusterKeyChange()":
                                                if vd["columnName"] != "clusterKeyChange":
                                                    for field in link.schema.fields:
                                                        if field.configuration.name == vd["columnName"]:
                                                            field.configuration.app_data["cluster_key_change"] = True
                                            elif vd["parsedExpression"] == "KeyChange()":
                                                if vd["columnName"] != "keyChange":
                                                    for field in link.schema.fields:
                                                        if field.configuration.name == vd["columnName"]:
                                                            field.configuration.app_data["key_change"] = True

    def populate_pivot_schema(self, pipeline_model: models.Pipeline) -> None:
        for link in self.dag.links():
            if hasattr(link.src, "configuration"):
                if link.src.configuration.op_name == "PxPivot":
                    for node_model in pipeline_model.nodes:
                        if hasattr(node_model, "op") and node_model.op == "PxPivot":
                            if node_model.app_data.ui_data.label == link.src.label:
                                for output in node_model.outputs:
                                    if hasattr(output, "parameters") and output.parameters and "valueDerivation" in output.parameters:
                                        for vd in output.parameters["valueDerivation"]:
                                            if "parsedExpression" in vd and vd["parsedExpression"].startswith("Pivot"):
                                                for field in link.schema.fields:
                                                    if field.configuration.name == vd["columnName"]:
                                                        pivot_prop = vd["parsedExpression"].split("(")[-1].strip(")")
                                                        if pivot_prop != vd["columnName"]:
                                                            field.configuration.app_data["pivot_property"] = pivot_prop

    def populate_change_capture_schema(self, pipeline_model: models.Pipeline) -> None:
        for link in self.dag.links():
            if hasattr(link.src, "configuration"):
                if link.src.configuration.op_name == "PxChangeCapture":
                    for node_model in pipeline_model.nodes:
                        if hasattr(node_model, "op") and node_model.op == "PxChangeCapture":
                            if node_model.app_data.ui_data.label == link.src.label:
                                for output in node_model.outputs:
                                    if hasattr(output, "parameters") and output.parameters and "valueDerivation" in output.parameters:
                                        for vd in output.parameters["valueDerivation"]:
                                            if vd["parsedExpression"] == "ChangeCode()":
                                                for field in link.schema.fields:
                                                    if field.configuration.name == vd["columnName"]:
                                                        field.configuration.app_data["change_code"] = True

    def populate_transformer_schema(self, pipeline_model: models.Pipeline) -> None:
        for link in self.dag.links():
            if hasattr(link.src, "configuration"):
                if link.src.configuration.op_name == "CTransformerStage":
                    for node_model in pipeline_model.nodes:
                        if hasattr(node_model, "op") and node_model.op == "CTransformerStage":
                            if node_model.app_data.ui_data.label == link.src.label:
                                for output in node_model.outputs:
                                    schema = None
                                    for schema_id, schema_model in self.schemas.items():
                                        if output.schema_ref == schema_id:
                                            schema = schema_model
                                    if hasattr(output, "parameters") and output.parameters and "valueDerivation" in output.parameters:
                                        for vd in output.parameters["valueDerivation"]:
                                            for field in schema.fields:
                                                if field.configuration.name == vd["columnName"]:
                                                    if "parsedExpression" in vd:
                                                        field.configuration.app_data["derivation"] = vd["parsedExpression"]
                                                    if "sourceColumn" in vd:
                                                        field.configuration.metadata.source_field_id = vd["sourceColumn"]
                                        del output.parameters["valueDerivation"]

    def populate_rest_schema(self) -> None:
        for link in self.dag.links():
            if hasattr(link.src, "configuration"):
                if link.src.configuration.op_name == "PxRest":
                    if link.schema:
                        for field in link.schema.fields:
                            if field.configuration.metadata.description:
                                field.configuration.app_data["derivation"] = field.configuration.metadata.description

    # def process_comments(self, pipeline_model: models.Pipeline) -> None:
    #     for comment in pipeline_model.app_data.ui_data.comments:
    #         if comment.content_type == "WYSIWYG":
    #             if comment.associated_id_refs:
    #                 for comment_ref in comment.associated_id_refs:
    #                     if comment_ref.node_ref in self.id_to_dag_node:
    #                         dag_node = self.id_to_dag_node[comment_ref.node_ref]
    #                         new_comment = StyledComment(dag=self.dag, configuration=comment)
    #                         new_comment.connect_output_to(dag_node)
    #             else:
    #                 new_comment = StyledComment(dag=self.dag, configuration=comment)
    #                 self.dag.add_node(new_comment)
    #         else:
    #             if comment.associated_id_refs:
    #                 for comment_ref in comment.associated_id_refs:
    #                     if comment_ref.node_ref in self.id_to_dag_node:
    #                         dag_node = self.id_to_dag_node[comment_ref.node_ref]
    #                         new_comment = MarkdownComment(dag=self.dag, content=comment.content)
    #                         new_comment.connect_output_to(dag_node)
    #             else:
    #                 new_comment = MarkdownComment(dag=self.dag, content=comment.content)
    #                 self.dag.add_node(new_comment)

    def link_up_node(
        self,
        dag_node: Node,
        node_model: models.Node,
        dag: DAG = None,
        id_to_dag_node: dict = None,
    ) -> None:
        """Links up the given node with other nodes in the DAG based on the links represented in its JSON model"""

        def create_link(link_model: models.NodeLink, src: Node, dest: Node, schema_ref: str | None) -> Link:
            """Creates a Link object from the given link model and source and destination nodes"""

            link = Link(dag=self.dag, name=link_model.link_name)
            link.src = src
            link.dest = dest

            if schema_ref:
                if schema_ref not in self.schemas:
                    raise KeyError(f"Could not find schema {schema_ref} referenced by node output")
                link.schema = self.schemas[schema_ref]

            type_attr = link_model.type_attr
            if type_attr is None:
                try:
                    type_attr = link_model.app_data.datastage["link_type"]
                except (AttributeError, KeyError):
                    type_attr = type_attr

            match type_attr:
                case "PRIMARY":
                    link.primary()
                case "REFERENCE":
                    link.reference()
                case "REJECT":
                    link.reject()
                case other:
                    raise ValueError(f"Unsupported link type: {other}")

            try:
                partner_link_out = link_model.app_data.datastage["partner_link_out"]
                link.maps_from_link = partner_link_out
            except (AttributeError, KeyError):
                pass
            try:
                partner_link_in = link_model.app_data.datastage["partner_link_in"]
                link.maps_to_link = partner_link_in
            except (AttributeError, KeyError):
                pass

            return link

        if hasattr(node_model, "inputs"):
            for input_port_model in node_model.inputs or []:
                if isinstance(input_port_model, tuple) and [] in input_port_model:
                    continue

                if not hasattr(input_port_model, "links"):
                    input_port_model = input_port_model[1][0]

                for link_model in input_port_model.links or []:
                    src_node_id = link_model.node_id_ref
                    if id_to_dag_node:
                        src_node = id_to_dag_node[src_node_id]
                    else:
                        src_node = self.id_to_dag_node[src_node_id]

                    if dag and src_node not in dag.adj:
                        raise KeyError(f"Could not find node {src_node}")
                    elif src_node not in self.dag.adj:
                        raise KeyError(f"Could not find node {src_node}")

                    if dag:
                        dag.adj[src_node].setdefault(dag_node, [])
                    else:
                        self.dag.adj[src_node].setdefault(dag_node, [])

                    link = create_link(link_model, src_node, dag_node, input_port_model.schema_ref)
                    if dag:
                        dag.adj[src_node][dag_node].append(link)
                    else:
                        self.dag.adj[src_node][dag_node].append(link)

        if dag and id_to_dag_node:
            return dag, id_to_dag_node

        # Commented out because there are never any output links
        # if hasattr(node_model, "outputs"):
        #     for output_port_model in node_model.outputs or []:
        #         for link_model in output_port_model.links or []:
        #             dest_node_id = link_model.node_id_ref
        #             dest_node = self.id_to_dag_node[dest_node_id]

        #             assert dag_node in self.dag.adj
        #             self.dag.adj[dag_node].setdefault(dest_node, [])

        #             link = create_link(link_model, dag_node, dest_node)
        #             self.dag.adj[dag_node][dest_node].append(link)

    def generate_node(self, node_model: models.Node) -> Node:
        """Given a node JSON model, pulls various information and returns a StageNode"""
        if isinstance(node_model, models.ExecutionNode):
            model_stage_node = self.generate_execution_node(node_model)
        elif isinstance(node_model, models.BindingEntryNode):
            model_stage_node = self.generate_binding_entry_node(node_model)
        elif isinstance(node_model, models.BindingExitNode):
            model_stage_node = self.generate_binding_exit_node(node_model)
        elif isinstance(node_model, models.ModelNode):
            model_stage_node = self.generate_model_node(node_model)
        elif isinstance(node_model, models.Supernode):
            raise TypeError("SuperNodes are not supported.")
            # model_stage_node = self.generate_super_node(node_model)
        else:
            raise TypeError(f"Unknown node model: {type(node_model).__class__.__name__}")

        # Pull label from JSON and assign it to the DAG node
        try:
            label = node_model.app_data.ui_data.label
        except AttributeError:
            label = None
        model_stage_node.label = label

        return model_stage_node

    def get_schema_ref_from_node(self, node: models.Node) -> models.RecordSchema | None:
        schema_ref = None
        out_ports = node.outputs if hasattr(node, "outputs") else []

        for out_port in out_ports:
            if hasattr(out_port, "schema_ref") and out_port.schema_ref:
                if schema_ref:
                    raise ValueError(f"Found multiple output ports with schema ref in node {node.id}")
                schema_ref = out_port.schema_ref
        return schema_ref

    def generate_execution_node(self, node_model: models.ExecutionNode) -> Node:
        if hasattr(node_model, "parameters") and node_model.parameters:
            model_stage_props = node_model.parameters.copy()
        else:
            model_stage_props = {}

        lookup_properties = []
        transformer_properties = []
        cff_properties = []
        for out_ in node_model.outputs or []:
            if out_.parameters:
                if node_model.op == "CTransformerStage":
                    if "outputName" in out_.parameters:
                        transformer_properties.append((out_.parameters, out_.parameters["outputName"]))
                    else:
                        output_id = out_.id
                        link_name = None
                        for pipeline in self.flow_model.pipelines:
                            for node in pipeline.nodes:
                                if hasattr(node, "inputs") and node.inputs:
                                    for input in node.inputs:
                                        if input.links:
                                            for link in input.links:
                                                if link.port_id_ref == output_id:
                                                    link_name = link.link_name
                        transformer_properties.append((out_.parameters, link_name))
                elif node_model.op == "PxCFF":
                    if "outputName" in out_.parameters:
                        cff_properties.append((out_.parameters, out_.parameters["outputName"]))
                    else:
                        output_id = out_.id
                        link_name = None
                        for pipeline in self.flow_model.pipelines:
                            for node in pipeline.nodes:
                                if hasattr(node, "inputs") and node.inputs:
                                    for input in node.inputs:
                                        if input.links:
                                            for link in input.links:
                                                if link.port_id_ref == output_id:
                                                    link_name = link.link_name
                        cff_properties.append((out_.parameters, link_name))
                if node_model.op != "JavaStagePX":
                    model_stage_props.update(out_.parameters)

        if node_model.inputs:
            for in_ in node_model.inputs:
                if in_.parameters:
                    if node_model.op == "PxLookup":
                        for link in in_.links:
                            if link.type_attr != "PRIMARY":
                                lookup_properties.append((in_.parameters, link.link_name))
                            else:
                                if "runtime_column_propagation" in in_.parameters:
                                    model_stage_props["runtime_column_propagation"] = in_.parameters["runtime_column_propagation"]
                    elif node_model.op != "JavaStagePX":
                        if "runtime_column_propagation" in in_.parameters:
                            for link in in_.links:
                                self.rcp[link.link_name] = bool(in_.parameters["runtime_column_propagation"])
                        model_stage_props.update(in_.parameters)

        if node_model.op == "CTransformerStage":
            # deal with string to int/bool
            if "LoopVariables" in model_stage_props:
                loop_variables = model_stage_props["LoopVariables"]
                for loop_var in loop_variables:
                    if "Precision" in loop_var:
                        loop_var["Precision"] = int(loop_var["Precision"])
                    if "Extended" in loop_var:
                        loop_var["Extended"] = True if str(loop_var["Extended"]).lower() == "true" or loop_var["Extended"] else False
                    if "Scale" in loop_var:
                        loop_var["Scale"] = int(loop_var["Scale"])
                model_stage_props["LoopVariables"] = [LoopVariable.model_construct(**loop_var) for loop_var in loop_variables]
            if "StageVariables" in model_stage_props:
                stage_variables = model_stage_props["StageVariables"]
                for stage_var in stage_variables:
                    if "Precision" in stage_var:
                        stage_var["Precision"] = int(stage_var["Precision"])
                    if "Extended" in stage_var:
                        stage_var["Extended"] = True if stage_var["Extended"] or str(stage_var["Extended"]).lower() == "true" else False
                    if "Scale" in stage_var:
                        stage_var["Scale"] = int(stage_var["Scale"])
                model_stage_props["StageVariables"] = [StageVariable.model_construct(**stage_var) for stage_var in stage_variables]
            if "Triggers" in model_stage_props:
                # deal with arguments here
                triggers = model_stage_props["Triggers"]
                for trigger in triggers:
                    arguments = []
                    for i in range(1, len(trigger)):
                        if f"Argument{i}" in trigger:
                            arguments.append(trigger[f"Argument{i}"])
                    trigger["arguments"] = arguments
                model_stage_props["Triggers"] = [Trigger.model_construct(**trigger) for trigger in triggers]
            # deal with constraints
            if len(transformer_properties) > 0:
                constraints = []
                for constraint_properties, link_name in transformer_properties:
                    constraint_props = {}
                    if "TransformerConstraint" in constraint_properties:
                        constraint_props["constraint"] = constraint_properties["TransformerConstraint"]
                    if "Reject" in constraint_properties:
                        constraint_props["otherwise_log"] = constraint_properties["Reject"]
                    if "RowLimit" in constraint_properties:
                        constraint_props["abort_after_rows"] = constraint_properties["RowLimit"]
                    if not len(constraint_props):
                        continue
                    constraint_props["output_name"] = link_name
                    constraints.append(Constraint.model_construct(**constraint_props))
                if len(constraints) > 0:
                    if "Reject" in model_stage_props:
                        del model_stage_props["Reject"]
                    if "RowLimit" in model_stage_props:
                        del model_stage_props["RowLimit"]
                model_stage_props["TransformerConstraint"] = constraints

        if node_model.op == "PxCFF":
            schemas = self.flow_model.schemas
            if (
                hasattr(node_model, "app_data")
                and node_model.app_data
                and hasattr(node_model.app_data, "datastage")
                and node_model.app_data.datastage
                and "stage_records" in node_model.app_data.datastage
            ):
                stage_records = node_model.app_data.datastage["stage_records"]
                records = []
                record_ids = []
                for stage_record in stage_records:
                    if "schema_ref" in stage_record:
                        for schema in schemas:
                            if schema.id == stage_record["schema_ref"]:
                                columns = []
                                for field in schema.fields:
                                    field_data = field.model_dump(exclude_none=True)
                                    properties = {
                                        "name": (field_data["name"] if "name" in field_data else None),
                                        "type": (field_data["type"] if "type" in field_data else None),
                                        "nullable": (field_data["nullable"] if "nullable" in field_data else None),
                                        **field_data["metadata"],
                                        **field_data["app_data"],
                                    }
                                    column = Column.from_dict(properties)
                                    columns.append(column)
                                record_name = stage_record["record_name"]
                                record = Record(name=record_name, columns=columns)
                                records.append(record)
                    record_id = RecordID(**stage_record)
                    record_ids.append(record_id)
                    model_stage_props["records"] = records
                    model_stage_props["records_id"] = record_ids
            if node_model.outputs:
                output_columns = []
                for output in node_model.outputs:
                    output_column = []
                    link_name = None
                    if output.schema_ref:
                        for schema in schemas:
                            if schema.id == output.schema_ref:
                                for field in schema.fields:
                                    output_column.append(field.name)
                    for pipeline in self.flow_model.pipelines:
                        for node in pipeline.nodes:
                            if hasattr(node, "inputs") and node.inputs:
                                for input in node.inputs:
                                    if input.links:
                                        for link in input.links:
                                            if link.port_id_ref == output.id:
                                                link_name = link.link_name
                    if link_name:
                        output_col_obj = OutputColumns(output_name=link_name, output_columns=output_column)
                        output_columns.append(output_col_obj)
                model_stage_props["output_columns"] = output_columns
            if len(cff_properties) > 0:
                constraints = []
                for constraint_properties, link_name in cff_properties:
                    constraint_props = {}
                    if "predicate" in constraint_properties:
                        constraint_props["constraint"] = constraint_properties["predicate"]
                    if not (len(constraint_props)):
                        continue
                    constraint_props["output_name"] = link_name
                    constraints.append(CFFConstraint.model_construct(**constraint_props))
                model_stage_props["predicate"] = constraints

        if node_model.op == "PxRest":
            if "requests" in model_stage_props:
                requests = []
                for request in model_stage_props["requests"]:
                    method = request["endpoint"]["method"] if "endpoint" in request and "method" in request["endpoint"] else None
                    use_expression_url = (
                        request["endpoint"]["expression_endpoint_on"]
                        if "endpoint" in request and "expression_endpoint_on" in request["endpoint"]
                        else False
                    )
                    url = None
                    auth = None
                    request_info = None
                    response = None
                    settings = None
                    control = None
                    if "endpoint" in request and "url" in request["endpoint"]:
                        if isinstance(request["endpoint"]["url"], dict) and "expression" in request["endpoint"]["url"]:
                            url = request["endpoint"]["url"]["expression"]
                        else:
                            url = request["endpoint"]["url"]
                    if "authentication" in request:
                        authentication = request["authentication"]
                        if "inherit" in authentication and authentication["inherit"]:
                            auth = Authentication(same_config=True)
                        else:
                            if "body" in authentication:
                                auth_props = {}
                                for key, value in authentication["body"].items():
                                    if isinstance(value, dict) and "expression" in value:
                                        auth_props[key] = value["expression"]
                                    else:
                                        auth_props[key] = value
                                auth = Authentication(**auth_props)
                    if "request" in request:
                        request_dict = request["request"]
                        if "inherit" in request_dict and request_dict["inherit"]:
                            request_info = RequestInfo(same_config=True)
                        else:
                            headers = []
                            cookies = []
                            parameters = []
                            body = None
                            additional_headers_on = False
                            addition_headers = None
                            if "headers" in request_dict:
                                for header in request_dict["headers"]:
                                    header_props = {}
                                    for key, value in header.items():
                                        if isinstance(value, dict) and "expression" in value:
                                            header_props[key] = value["expression"]
                                        else:
                                            header_props[key] = value
                                    new_header = Header(**header_props)
                                    headers.append(new_header)
                            if "cookies" in request_dict:
                                for cookie in request_dict["cookies"]:
                                    cookie_props = {}
                                    for key, value in cookie.items():
                                        if isinstance(value, dict) and "expression" in value:
                                            cookie_props[key] = value["expression"]
                                        else:
                                            cookie_props[key] = value
                                    new_cookie = Cookie(**cookie_props)
                                    cookies.append(new_cookie)
                            if "params" in request_dict:
                                for param in request_dict["params"]:
                                    param_props = {}
                                    for key, value in param.items():
                                        if isinstance(value, dict) and "expression" in value:
                                            param_props[key] = value["expression"]
                                        else:
                                            param_props[key] = value
                                    new_param = Parameter(**param_props)
                                    parameters.append(new_param)
                            if "body" in request_dict:
                                body_props = {}
                                if "content" in request_dict["body"]:
                                    type = request_dict["body"]["content"]["type"] if "type" in request_dict["body"]["content"] else None
                                    body_props["type"] = type
                                    if type == "BINARY":
                                        if "data" in request_dict["body"]["content"]:
                                            if (
                                                isinstance(
                                                    request_dict["body"]["content"]["data"],
                                                    dict,
                                                )
                                                and "expression" in request_dict["body"]["content"]["data"]
                                            ):
                                                if "source" in request_dict["body"]["content"]:
                                                    if request_dict["body"]["content"]["source"] == "DATA":
                                                        body_props["binary_data"] = request_dict["body"]["content"]["data"]["expression"]
                                                    elif request_dict["body"]["content"]["source"] == "FILE":
                                                        body_props["file_path"] = request_dict["body"]["content"]["data"]["expression"]
                                    if type == "RAW":
                                        if "data" in request_dict["body"]["content"]:
                                            if (
                                                isinstance(
                                                    request_dict["body"]["content"]["data"],
                                                    dict,
                                                )
                                                and "expression" in request_dict["body"]["content"]["data"]
                                            ):
                                                if "source" in request_dict["body"]["content"]:
                                                    if request_dict["body"]["content"]["source"] == "TEXT":
                                                        body_props["raw_text"] = request_dict["body"]["content"]["data"]["expression"]
                                                    elif request_dict["body"]["content"]["source"] == "FILE":
                                                        body_props["file_path"] = request_dict["body"]["content"]["data"]["expression"]
                                    if type == "FORM_DATA":
                                        if "data" in request_dict["body"]["content"]:
                                            if isinstance(
                                                request_dict["body"]["content"]["data"],
                                                list,
                                            ):
                                                form_datas = []
                                                for form_data in request_dict["body"]["content"]["data"]:
                                                    data_props = {}
                                                    for key, value in form_data.items():
                                                        if isinstance(value, dict) and "expression" in value:
                                                            data_props[key] = value["expression"]
                                                        else:
                                                            data_props[key] = value
                                                    new_form_data = FormData(**data_props)
                                                    form_datas.append(new_form_data)
                                                body_props["form_data"] = form_datas
                                    if type == "X_WWW_FORM_URLENCODED":
                                        if "data" in request_dict["body"]["content"]:
                                            if isinstance(
                                                request_dict["body"]["content"]["data"],
                                                list,
                                            ):
                                                form_datas = []
                                                for form_data in request_dict["body"]["content"]["data"]:
                                                    data_props = {}
                                                    for key, value in form_data.items():
                                                        if isinstance(value, dict) and "expression" in value:
                                                            data_props[key] = value["expression"]
                                                        else:
                                                            data_props[key] = value
                                                    new_form_data = FormURLEncodedData(**data_props)
                                                    form_datas.append(new_form_data)
                                                body_props["form_urlencoded_data"] = form_datas
                                    if "content_type" in request_dict["body"]["content"]:
                                        body_props["content_type"] = request_dict["body"]["content"]["content_type"]
                                    if "charset" in request_dict["body"]["content"]:
                                        body_props["charset"] = request_dict["body"]["content"]["charset"]
                                    if "expression_filePath_on" in request_dict:
                                        body_props["expression_filePath_on"] = request_dict["body"]["content"]["expression_filePath_on"]
                                    if "expression_text_on" in request_dict:
                                        if "source" in request_dict["body"]["content"]:
                                            if request_dict["body"]["content"]["source"] == "DATA":
                                                body_props["use_expression_data"] = request_dict["body"]["content"]["expression_text_on"]
                                            if request_dict["body"]["content"]["source"] == "TEXT":
                                                body_props["use_expression_text"] = request_dict["body"]["content"]["expression_text_on"]
                                body = Body(**body_props)
                                if "additional_headers_on" in request_dict["body"]:
                                    additional_headers_on = request_dict["body"]["additional_headers_on"]
                            if "additional_headers" in request_dict:
                                if (
                                    isinstance(request_dict["additional_headers"], dict)
                                    and "expression" in request_dict["additional_headers"]
                                ):
                                    addition_headers = request_dict["additional_headers"]["expression"]
                                else:
                                    addition_headers = request_dict["additional_headers"]
                            request_info = RequestInfo(
                                query_parameters=parameters,
                                custom_headers=headers,
                                custom_cookies=cookies,
                                body=body,
                                additional_headers_on=additional_headers_on,
                                additional_headers=addition_headers,
                            )
                    if "response" in request:
                        response_dict = request["response"]
                        if "inherit" in response_dict and response_dict["inherit"]:
                            response = Response(same_config=True)
                        else:
                            response_props = {}
                            if "body" in response_dict:
                                if "content_type" in response_dict["body"]:
                                    response_props["content_type"] = response_dict["body"]["content_type"]
                                if "charset" in response_dict["body"]:
                                    response_props["charset"] = response_dict["body"]["charset"]
                                if "content" in response_dict["body"]:
                                    if "target" in response_dict["body"]["content"]:
                                        response_props["target"] = response_dict["body"]["content"]["target"]
                                    if "file_path" in response_dict["body"]["content"]:
                                        if (
                                            isinstance(
                                                response_dict["body"]["content"]["file_path"],
                                                dict,
                                            )
                                            and "expression" in response_dict["body"]["content"]["file_path"]
                                        ):
                                            response_props["file_path"] = response_dict["body"]["content"]["file_path"]["expression"]
                                        else:
                                            response_props["file_path"] = response_dict["body"]["content"]["file_path"]
                                if "expression_filePath_on" in response_dict["body"]:
                                    response_props["expression_filePath_on"] = response_dict["body"]["expression_filePath_on"]
                            response = Response(**response_props)
                    if "settings" in request:
                        if "inherit" in request["settings"] and request["settings"]["inherit"]:
                            settings = Settings(same_config=True)
                        else:
                            settings_props = {}
                            for key, val in request["settings"].items():
                                if key == "server_cert":
                                    server_cert_props = {}
                                    for serv_key, serv_prop in request["settings"][key].items():
                                        if isinstance(serv_prop, dict) and "expression" in serv_prop:
                                            server_cert_props[serv_key] = serv_prop["expression"]
                                        else:
                                            server_cert_props[serv_key] = serv_prop
                                    settings_props["server_cert"] = server_cert_props
                                elif key == "client_cert":
                                    client_cert_props = {}
                                    for client_key, client_prop in request["settings"][key].items():
                                        if isinstance(client_prop, dict) and "expression" in client_prop:
                                            client_cert_props[client_key] = client_prop["expression"]
                                        else:
                                            client_cert_props[client_key] = client_prop
                                    settings_props["client_cert"] = client_cert_props
                                elif key in [
                                    "client_certificate",
                                    "server_certificate",
                                ]:
                                    continue
                                else:
                                    settings_props[key] = val
                            settings = Settings(**settings_props)
                    if "control" in request:
                        if "inherit" in request["control"] and request["control"]["inherit"]:
                            control = Control(same_config=True)
                        else:
                            control_props = {}
                            if "write_output_per_iteration" in request["control"] and request["control"]["write_output_per_iteration"]:
                                request["control"]["output_type"] = "PER_ITERATION"
                            for key, value in request["control"].items():
                                if isinstance(value, dict) and "expression" in value:
                                    control_props[key] = value["expression"]
                                else:
                                    control_props[key] = value
                            control = Control(**control_props)

                    new_request = Request(
                        method=method,
                        url=url,
                        use_expression_url=use_expression_url,
                        authentication=auth,
                        request=request_info,
                        response=response,
                        settings=settings,
                        control=control,
                    )
                    requests.append(new_request)
                model_stage_props["requests"] = requests
            if "variables" in model_stage_props:
                variables = []
                for var in model_stage_props["variables"]:
                    variable = {}
                    for key, val in var.items():
                        if isinstance(val, dict) and "expression" in val:
                            variable[key] = val["expression"]
                        else:
                            variable[key] = val
                    new_var = Variable(**variable)
                    variables.append(new_var)
                model_stage_props["variables"] = variables

        if node_model.op == "PxJoin" and "InputlinkOrderingList" not in model_stage_props:
            if (
                hasattr(node_model, "app_data")
                and node_model.app_data is not None
                and hasattr(node_model.app_data, "datastage")
                and node_model.app_data.datastage is not None
                and "inputs_order" in node_model.app_data.datastage
                and node_model.app_data.datastage["inputs_order"] is not None
            ):
                inputs_order = node_model.app_data.datastage["inputs_order"].split("|")
                ordering_list = []
                for i in range(len(inputs_order)):
                    link_name = None
                    for input in node_model.inputs:
                        if input.id == inputs_order[i]:
                            link_name = input.links[0].link_name
                    if link_name:
                        if i == 0:
                            ordering_list.append({"link_label": "Left", "link_name": link_name})
                        elif i == len(inputs_order) - 1:
                            ordering_list.append({"link_label": "Right", "link_name": link_name})
                        else:
                            key = f"Intermediate {i}"
                            ordering_list.append({"link_label": key, "link_name": link_name})
                model_stage_props["InputlinkOrderingList"] = ordering_list

        if node_model.op == "PxLookup":
            if "InputlinkOrderingList" not in model_stage_props:
                if (
                    hasattr(node_model, "app_data")
                    and node_model.app_data is not None
                    and hasattr(node_model.app_data, "datastage")
                    and node_model.app_data.datastage is not None
                    and "inputs_order" in node_model.app_data.datastage
                    and node_model.app_data.datastage["inputs_order"] is not None
                ):
                    inputs_order = node_model.app_data.datastage["inputs_order"].split("|")
                    ordering_list = []
                    primary_first = False
                    for input in node_model.inputs:
                        if len(inputs_order) and input.id == inputs_order[0]:
                            if input.links[0].type_attr == "PRIMARY":
                                primary_first = True
                    for i in range(len(inputs_order)):
                        link_name = None
                        for input in node_model.inputs:
                            if input.id == inputs_order[i]:
                                link_name = input.links[0].link_name
                        if link_name:
                            if primary_first:
                                if i == 0:
                                    ordering_list.append(
                                        {
                                            "link_label": "Primary",
                                            "link_name": link_name,
                                        }
                                    )
                                else:
                                    ordering_list.append(
                                        {
                                            "link_label": f"Lookup {i}",
                                            "link_name": link_name,
                                        }
                                    )
                            else:
                                if i == len(inputs_order) - 1:
                                    ordering_list.append(
                                        {
                                            "link_label": "Primary",
                                            "link_name": link_name,
                                        }
                                    )
                                else:
                                    ordering_list.append(
                                        {
                                            "link_label": f"Lookup {i + 1}",
                                            "link_name": link_name,
                                        }
                                    )

                    model_stage_props["InputlinkOrderingList"] = ordering_list
            if len(lookup_properties) > 0:
                lookup_derivations = []
                for lookup_derivation, link_name in lookup_properties:
                    ld_props = {}
                    if "conditionNotMet" in lookup_derivation:
                        ld_props["condition_not_met"] = lookup_derivation["conditionNotMet"]
                    if "lookupFail" in lookup_derivation:
                        ld_props["lookup_failure"] = lookup_derivation["lookupFail"]
                    if "condition" in lookup_derivation:
                        ld_props["condition"] = lookup_derivation["condition"]
                    if "Condition" in lookup_derivation:
                        ld_props["condition"] = lookup_derivation["Condition"]
                    ld_props["reference_link"] = link_name
                    if "lookupDerivation" in lookup_derivation and lookup_derivation["lookupDerivation"]:
                        derivations = []
                        for ld in lookup_derivation["lookupDerivation"]:
                            derivation = {}
                            if "columnName" in ld:
                                derivation["key_column"] = ld["columnName"]
                            if "keyType" in ld:
                                derivation["key_type"] = ld["keyType"]
                            if "parsedExpression" in ld:
                                if "Range" in ld["parsedExpression"]:
                                    derivation["key_expression"] = ld["parsedExpression"]
                                else:
                                    derivation["key_expression"] = ld["parsedExpression"].split(".")[-1]
                            derivations.append(derivation)
                        ld_props["derivations"] = derivations
                    ld_class = lookup.LookupDerivation.model_construct(**ld_props) if len(ld_props) else None
                    if ld_class:
                        # IDEALLY GET RID OF THIS
                        if not isinstance(ld_class.lookup_failure, str) and ld_class.lookup_failure:
                            ld_class.lookup_failure = ld_class.lookup_failure.value
                        if not isinstance(ld_class.condition_not_met, str) and ld_class.condition_not_met:
                            ld_class.condition_not_met = ld_class.condition_not_met.value
                        lookup_derivations.append(ld_class)
                model_stage_props["lookup_derivation"] = lookup_derivations

        if node_model.op == "PxChangeCapture" and "InputlinkOrderingList" not in model_stage_props:
            if (
                hasattr(node_model, "app_data")
                and node_model.app_data is not None
                and hasattr(node_model.app_data, "datastage")
                and node_model.app_data.datastage is not None
                and "inputs_order" in node_model.app_data.datastage
                and node_model.app_data.datastage["inputs_order"] is not None
            ):
                inputs_order = node_model.app_data.datastage["inputs_order"].split("|")
                ordering_list = []
                for i in range(len(inputs_order)):
                    link_name = None
                    for input in node_model.inputs:
                        if input.id == inputs_order[i]:
                            link_name = input.links[0].link_name
                    if link_name:
                        if i == 0:
                            ordering_list.append({"link_label": "Before", "link_name": link_name})
                        else:
                            ordering_list.append({"link_label": "After", "link_name": link_name})
                model_stage_props["InputlinkOrderingList"] = ordering_list

        if node_model.op == "PxGeneric":
            if "InputlinkOrderingList" not in model_stage_props:
                if (
                    hasattr(node_model, "app_data")
                    and node_model.app_data is not None
                    and hasattr(node_model.app_data, "datastage")
                    and node_model.app_data.datastage is not None
                    and "inputs_order" in node_model.app_data.datastage
                    and node_model.app_data.datastage["inputs_order"] is not None
                ):
                    inputs_order = node_model.app_data.datastage["inputs_order"].split("|")
                    ordering_list = []
                    for i in range(len(inputs_order)):
                        link_name = None
                        for input in node_model.inputs:
                            if input.id == inputs_order[i]:
                                link_name = input.links[0].link_name
                        if link_name:
                            ordering_list.append({"link_label": f"{i}", "link_name": link_name})
                    model_stage_props["InputlinkOrderingList"] = ordering_list

            if "OutputlinkOrderingList" not in model_stage_props:
                if (
                    hasattr(node_model, "app_data")
                    and node_model.app_data is not None
                    and hasattr(node_model.app_data, "datastage")
                    and node_model.app_data.datastage is not None
                    and "outputs_order" in node_model.app_data.datastage
                    and node_model.app_data.datastage["outputs_order"] is not None
                ):
                    outputs_order = node_model.app_data.datastage["outputs_order"].split("|")
                    ordering_list = []
                    for i in range(len(outputs_order)):
                        link_name = None
                        for output in node_model.outputs:
                            if output.id == outputs_order[i] and output.links:
                                # TODO: outputs don't have link_name, can grab from corresponding input
                                # link, but it would require passing extra data to the method
                                link_name = output.links[0].link_name
                        if link_name:
                            ordering_list.append({"link_label": f"{i}", "link_name": link_name})
                    model_stage_props["OutputlinkOrderingList"] = ordering_list

        if node_model.op == "PxChecksum" and "valueDerivation" in model_stage_props:
            for value_deriv in model_stage_props["valueDerivation"]:
                if "parsedExpression" in value_deriv and value_deriv["parsedExpression"] == "Checksum()":
                    if "columnName" in value_deriv:
                        model_stage_props["checksum_name"] = value_deriv["columnName"]

        if node_model.op in ["PxJoin", "PxSort", "PxFunnel", "PxAggregator", "PxRemDup", "PxFilter"] and "keyColsPart" in model_stage_props:
            del model_stage_props["keyColsPart"]

        if node_model.op in ["PxJoin", "PxSort", "PxFunnel", "PxAggregator", "PxRemDup", "PxFilter"] and "keyColsColl" in model_stage_props:
            del model_stage_props["keyColsColl"]

        if node_model.op in ["PxJoin", "PxSort", "PxFunnel", "PxAggregator", "PxRemDup", "PxFilter"] and "keyColsNone" in model_stage_props:
            del model_stage_props["keyColsNone"]

        if "runtime_column_propagation" in model_stage_props:
            if model_stage_props["runtime_column_propagation"] in [0, 1]:
                model_stage_props["runtime_column_propagation"] = bool(model_stage_props["runtime_column_propagation"])

        if hasattr(node_model.app_data, "datastage") and (
            "custom_op_asset_id" in node_model.app_data.datastage and "custom_optype" in node_model.app_data.datastage
        ):
            if node_model.app_data.datastage["custom_optype"] == 0:
                return self.create_custom_stage_node(node_model, model_stage_props)
            elif node_model.app_data.datastage["custom_optype"] == 1:
                return self.create_build_stage_node(node_model, model_stage_props)
            elif node_model.app_data.datastage["custom_optype"] == 2:
                return self.create_wrapped_stage_node(node_model, model_stage_props)

        model_stage_node = self.create_stage_node(node_model, model_stage_props)
        return model_stage_node

    def generate_binding_entry_node(self, node_model: models.BindingEntryNode) -> Node:
        if hasattr(node_model, "parameters") and node_model.parameters:
            model_stage_props = node_model.parameters.copy()
        else:
            model_stage_props = {}

        for out_ in node_model.outputs or []:
            if out_.parameters:
                model_stage_props.update(out_.parameters)

        if not node_model.op:
            return self.create_entry_node(node_model)

        if node_model.connection:
            model_stage_props.update(node_model.connection.properties)

        return self.create_stage_node(node_model, model_stage_props)

    def generate_binding_exit_node(self, node_model: models.BindingExitNode) -> Node:
        if hasattr(node_model, "parameters") and node_model.parameters:
            model_stage_props = node_model.parameters.copy()
        else:
            model_stage_props = {}

        for out_ in node_model.outputs or []:
            if out_.parameters:
                model_stage_props.update(out_.parameters)

        if not node_model.op:
            return self.create_exit_node(node_model)

        if node_model.inputs or []:
            for in_ in node_model.inputs:
                if in_.parameters:
                    model_stage_props.update(in_.parameters)

        if node_model.connection:
            model_stage_props.update(node_model.connection.properties)

        return self.create_stage_node(node_model, model_stage_props)

    def generate_model_node(self, model_node: models.ModelNode):
        raise NotImplementedError("Model node generation is not supported yet")

    # def generate_super_node(self, super_node: models.Supernode):
    #     if hasattr(super_node.subflow_ref, "url") and super_node.subflow_ref.url:
    #         # If the url is present, then the subflow must be external
    #         # Use a placeholder SuperNodeRef here so that we can import the subflow asset later
    #         name = super_node.subflow_ref.name
    #         node = SuperNodeRef(self.fc._dag, name)
    #         self.dag.add_node(node)
    #         return node

    #     # URL is not present, so the subflow must be local
    #     # First, find the pipeline model referenced by the ID
    #     pipeline_id_ref = super_node.subflow_ref.pipeline_id_ref
    #     pipeline: models.Pipeline | None = None
    #     for pipe in self.flow_model.pipelines:
    #         if pipe.id == pipeline_id_ref:
    #             pipeline = pipe
    #             break

    #     if not pipeline:
    #         raise ValueError(f"Could not find local subflow referenced by ID {pipeline_id_ref}")

    #     flow_model = models.Flow(
    #         pipelines=[pipeline],
    #         schemas=self.flow_model.schemas,
    #         primary_pipeline=pipeline_id_ref,
    #     )
    #     pipeline_dag_gen = DAGGenerator(flow_model)
    #     dag = pipeline_dag_gen.generate()._dag

    #     entry_nodes = filter(
    #         lambda node: isinstance(node, EntryNode),
    #         self.dag.nodes(),
    #     )
    #     exit_nodes = filter(
    #         lambda node: isinstance(node, ExitNode),
    #         self.dag.nodes(),
    #     )
    #     name = pipeline.name or super_node.app_data.ui_data.label
    #     node = SuperNode(
    #         parent_dag=self.dag,
    #         subflow_dag=dag,
    #         is_local=True,
    #         entry_nodes=entry_nodes,
    #         exit_nodes=exit_nodes,
    #         name=name,
    #     )

    #     self.dag.add_node(node)
    #     return node

    # def create_build_stage_node(self, node_model: models.ExecutionNode, model_stage_props: dict):
    #     model_stage_props["op_name"] = node_model.op
    #     build_stage_model = build_stage.model_construct(**model_stage_props)
    #     property_names = [value.alias for value in build_stage_model.model_fields.values()] + [
    #         key for key in build_stage_model.model_fields.keys()
    #     ]
    #     properties = {}
    #     for prop in model_stage_props:
    #         if prop not in property_names and prop not in [
    #             "showPartType",
    #             "showCollType",
    #             "showSortOptions",
    #             "conversion",
    #             "operator",
    #             "keyColsPart",
    #             "outputName",
    #             "outputcolProperties",
    #         ]:
    #             properties[prop] = model_stage_props[prop]
    #     build_stage_stage = BuildStageStage(
    #         dag=self.fc._dag,
    #         build_stage_asset=None,
    #         label=node_model.app_data.ui_data.label,
    #         configuration=build_stage_model,
    #     )
    #     build_stage_stage.properties = properties
    #     self.dag.add_node(build_stage_stage)
    #     return build_stage_stage

    # def create_wrapped_stage_node(self, node_model: models.ExecutionNode, model_stage_props: dict):
    #     model_stage_props["op_name"] = node_model.op
    #     wrapped_stage_model = wrapped_stage.model_construct(**model_stage_props)
    #     property_names = [value.alias for value in wrapped_stage_model.model_fields.values()] + [
    #         key for key in wrapped_stage_model.model_fields.keys()
    #     ]
    #     properties = {}
    #     for prop in model_stage_props:
    #         if prop not in property_names and prop not in [
    #             "showPartType",
    #             "showCollType",
    #             "showSortOptions",
    #             "conversion",
    #             "operator",
    #             "properties_order",
    #             "keyColsPart",
    #             "outputName",
    #             "outputcolProperties",
    #         ]:
    #             properties[prop] = model_stage_props[prop]
    #     wrapped_stage_stage = WrappedStageStage(
    #         dag=self.fc._dag,
    #         wrapped_stage_asset=None,
    #         label=node_model.app_data.ui_data.label,
    #         configuration=wrapped_stage_model,
    #     )
    #     wrapped_stage_stage.properties = properties
    #     self.dag.add_node(wrapped_stage_stage)
    #     return wrapped_stage_stage

    # def create_custom_stage_node(self, node_model: models.ExecutionNode, model_stage_props: dict):
    #     model_stage_props["op_name"] = node_model.op
    #     custom_stage_model = custom_stage.model_construct(**model_stage_props)
    #     property_names = [value.alias for value in custom_stage_model.model_fields.values()] + [
    #         key for key in custom_stage_model.model_fields.keys()
    #     ]
    #     properties = {}
    #     for prop in model_stage_props:
    #         if prop not in property_names and prop not in [
    #             "showPartType",
    #             "showCollType",
    #             "showSortOptions",
    #             "conversion",
    #             "operator",
    #             "preserve",
    #             "outputName",
    #             "outputcolProperties",
    #         ]:
    #             properties[prop] = model_stage_props[prop]
    #     custom_stage_stage = CustomStageStage(
    #         dag=self.fc._dag,
    #         custom_stage_asset=None,
    #         label=node_model.app_data.ui_data.label,
    #         configuration=custom_stage_model,
    #     )
    #     custom_stage_stage.properties = properties
    #     self.dag.add_node(custom_stage_stage)
    #     return custom_stage_stage

    # def create_entry_node(self, node_model: models.BindingEntryNode):
    #     """Creates an entry node generally used in subflows"""
    #     node = EntryNode(self.fc._dag)
    #     self.dag.add_node(node)
    #     return node

    # def create_exit_node(self, node_model: models.BindingExitNode):
    #     """Creates an exit node generally used in subflows"""
    #     node = ExitNode(self.fc._dag)
    #     self.dag.add_node(node)
    #     return node

    def create_stage_node(
        self,
        node_model: models.ExecutionNode | models.BindingEntryNode | models.BindingExitNode,
        model_stage_props: dict,
    ) -> StageNode:
        """Creates a model stage node from the given internal operator name and a dictionary of the model properties"""

        if node_model.op in OP_MAPPINGS:
            op_name = OP_MAPPINGS[node_model.op]
        else:
            op_name = "peek"
        connection_model_node = None
        java_library_model_node = None
        if (
            hasattr(node_model, "connection")
            and node_model.connection
            and (node_model.connection.ref and node_model.connection.ref != "local")
        ):
            if node_model.connection.name:
                conn_name = node_model.connection.name
            elif node_model.connection.connData and node_model.connection.connData["connectionName"]:
                conn_name = node_model.connection.connData["connectionName"]
            else:
                conn_name = node_model.app_data.ui_data.label + "_connection"

            connection_model_node = self.create_connection_model(op_name=op_name, asset_id=node_model.connection.ref, name=conn_name)
        if node_model.op == "JavaStagePX":
            files = node_model.parameters["classpath"].split(";")
            if len(files) > 1:
                secondary_files = files[1:]
            else:
                secondary_files = []

            if "asset_name" in node_model.parameters and "asset_id" in node_model.parameters:
                java_library_model_node = self.create_java_library_model(
                    name=node_model.parameters["asset_name"],
                    asset_id=node_model.parameters["asset_id"],
                    primary_file=files[0],
                    secondary_files=secondary_files,
                )

        stage_model = instantiate_stage_model(model_class=op_name, model_stage_props=model_stage_props)

        if connection_model_node:
            stage_model.connection = connection_model_node
        if java_library_model_node:
            stage_model.java_library = java_library_model_node

        stage_node = instantiate_stage_node(op_name, self.fc._dag, stage_model)

        self.dag.add_node(stage_node)
        return stage_node

    # TODO refactor to support creating connections
    def create_connection_model(self, op_name: str, asset_id: str, name: str):
        conn_props = {"asset_id": asset_id, "name": name}
        conn_model = instantiate_connection_model(op_name=op_name, conn_props=conn_props)

        return conn_model

    # def create_java_library_model(self, name: str, asset_id: str, primary_file: str, secondary_files: list[str]):
    #     jl_props = {
    #         "name": name,
    #         "asset_id": asset_id,
    #         "primary_file": primary_file,
    #         "secondary_files": secondary_files,
    #     }
    #     jl_model = instantiate_java_library_model(jl_props)
    #     return jl_model

    # def process_paramsets(self, paramset_model: models.ParamSet):
    #     return ParameterSet(name=paramset_model.name, asset_id=paramset_model.ref)

    # def process_localparams(self, local_params: list[dict]):
    #     if not local_params:
    #         return None
    #     return LocalParameters.from_dict({"parameters": local_params})

    # def process_message_handler(self, message_handler_id: str):
    #     return MessageHandler(name="referenced_message_handler", asset_id=message_handler_id)

    # def process_local_message_handler(self, local_messages: list[dict]):
    #     if not local_messages:
    #         return None
    #     return LocalMessageHandler(**{"messages": local_messages})

    # def process_runtime(self, app_data: models.AppData):
    #     if (app_data and hasattr(app_data, "datastage") and app_data.datastage) and (
    #         app_data.datastage.get("date_format")
    #         or app_data.datastage.get("timestamp_format")
    #         or app_data.datastage.get("decimal_separator")
    #         or app_data.datastage.get("time_format")
    #     ):
    #         format = Format(
    #             date_format=app_data.datastage.get("date_format"),
    #             timestamp_format=app_data.datastage.get("timestamp_format"),
    #             decimal_separator=app_data.datastage.get("decimal_separator"),
    #             time_format=app_data.datastage.get("time_format"),
    #         )
    #     else:
    #         format = None

    #     settings = None
    #     retention = (
    #         app_data.datastage.get("runArtifactRetentionJob")
    #         if app_data and hasattr(app_data, "datastage") and app_data.datastage
    #         else None
    #     )
    #     if retention:
    #         retention.get("retentionData")
    #     if retention and "days" in retention:
    #         settings = RuntimeSettings(retention_days=retention["days"])
    #     elif retention and "amount" in retention:
    #         settings = RuntimeSettings(retention_amount=retention["amount"])

    #     # TODO add ps/vs support
    #     return Runtime(runtime_settings=settings, format=format)


class ConnectionGenerator:
    def __init__(self, conn_json: dict):
        self.conn_json = conn_json
        self.conn_type = OP_MAPPINGS[DATASOURCE_MAPPINGS[conn_json["datasource_type"]]]

    def create_connection_model(self):
        conn_props = self.conn_json["properties"]
        conn_props["name"] = self.conn_json["name"]
        return instantiate_connection_model(op_name=self.conn_type, conn_props=conn_props)


# class JavaLibraryGenerator:
#     def __init__(self, jl_data: dict):
#         self.jl_data = jl_data

#     def create_java_library_model(self, output_path: str, write_to_output: bool):
#         entity = self.jl_data["entity"]
#         primary_jar_content = entity["primary_jar_content"]
#         primary_jar_name = list(entity["primary"])[0]
#         if write_to_output:
#             if not os.path.exists(output_path):
#                 _check_create_dir(os.getcwd() + "/attachments")
#                 output_path = os.getcwd()
#             else:
#                 _check_create_dir(output_path + "/attachments")
#             with open(Path(output_path) / "attachments" / (Path(primary_jar_name)), "wb") as f:
#                 f.write(base64.b64decode(primary_jar_content))
#         jl_props = {}
#         jl_props["primary_file"] = primary_jar_name
#         print(primary_jar_name)
#         secondary_files = []
#         for secondary_file in entity["secondary"]:
#             if write_to_output:
#                 with open(
#                     Path(output_path) / "attachments" / (Path(secondary_file["jar_file_name"])),
#                     "wb",
#                 ) as f:
#                     f.write(base64.b64decode(secondary_file["jar_file_content"]))
#             secondary_files.append(secondary_file["jar_file_name"])
#         jl_props["secondary_files"] = secondary_files
#         jl_props["name"] = self.jl_data["name"]
#         jl_props["description"] = self.jl_data["description"]
#         return instantiate_java_library_model(jl_props)


# class FunctionLibraryGenerator:
#     def __init__(self, fl_data: dict):
#         self.fl_data = fl_data

#     def create_function_library_model(self, output_path: str, write_to_output: bool):
#         entity = self.fl_data["entity"]
#         so_file_content = entity["so_file_content"]
#         function_library_name = list(entity["custom_function_name"])[0]
#         if write_to_output:
#             if not os.path.exists(output_path):
#                 _check_create_dir(os.getcwd() + "/attachments")
#                 output_path = os.getcwd()
#             else:
#                 _check_create_dir(output_path + "/attachments")
#             with open(Path(output_path) / "attachments" / (Path(function_library_name)), "wb") as f:
#                 f.write(base64.b64decode(so_file_content))
#         fl_props = {}
#         fl_props["library_path"] = function_library_name
#         fl_props["name"] = self.fl_data["name"]
#         fl_props["description"] = self.fl_data["description"]
#         return instantiate_function_library_model(fl_props)


# class MatchSpecificationGenerator:
#     def __init__(self, ms_data: dict):
#         self.ms_data = ms_data

#     def create_match_specification_model(self, output_path: str, write_to_output: bool):
#         entity = self.ms_data["entity"]
#         mat_json = entity["content"]["mat"]["MATCHSPEC"]
#         passes_json = entity["content"]["passes"]

#         if write_to_output:
#             if not os.path.exists(output_path):
#                 _check_create_dir(os.getcwd() + "/attachments")
#                 output_path = os.getcwd()
#             else:
#                 _check_create_dir(output_path + "/attachments")

#         match_properties = MatchSpecification.create_match_properties(self.ms_data["name"], mat_json, passes_json)

#         return instantiate_match_specification_model(match_properties)


# class JobGenerator:
#     def __init__(self, job_json: dict):
#         self.job_json = job_json

#     def create_job_model(self):
#         schedule_info = self.job_json.get("schedule_info")
#         start, end, repeat = None, None, None
#         if schedule_info:
#             start = schedule_info.get("startOn")
#             start = str(datetime.datetime.fromtimestamp(start / 1000)).replace(":", "-") if start else start
#             end = schedule_info.get("endOn")
#             end = str(datetime.datetime.fromtimestamp(end / 1000)).replace(":", "-") if end else end
#             repeat = schedule_info.get("repeat")
#         schedule = self.job_json.get("schedule")
#         schedule_obj = None
#         if schedule_info or schedule:
#             schedule_obj = Schedule(start=start, end=end, repeat=repeat, schedule=schedule)

#         parameter_sets = self.job_json.get("parameter_sets")
#         value_sets = {}
#         if parameter_sets:
#             for param_set in parameter_sets:
#                 value_sets[param_set.get("name")] = param_set.get("value_set")
#         runtime_parameters = RuntimeParameters(value_sets=value_sets)

#         job_settings = JobSettings(runtime_parameters=runtime_parameters, schedule=schedule_obj)

#         return job_settings



from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import PLANNING_ANALYTICS_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class AuthType(Enum):
    bearer = "bearer"
    cam_credentials = "cam_credentials"
    basic = "basic"


class PlanningAnalyticsConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "tm1odata"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    ds_host: str | None = Field(None, alias="_host")
    ds_port: str | None = Field(None, alias="_port")
    access_token: str = Field(None, alias="access_token")
    authentication_type: PLANNING_ANALYTICS_CONNECTION.AuthType | None = Field(None, alias="auth_type")
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    gateway_url: str | None = Field(None, alias="gateway_url")
    use_my_platform_login_credentials: bool | None = Field(False, alias="inherit_access_token")
    namespace: str = Field(None, alias="namespace")
    password: str = Field(None, alias="password")
    tm1_server_api_root_url: str = Field(None, alias="service_root")
    sl_client_cert: str | None = Field(None, alias="sl_client_cert")
    sl_client_private_key: str | None = Field(None, alias="sl_client_private_key")
    sl_connector_id: str | None = Field(None, alias="sl_connector_id")
    sl_endpoint_host: str | None = Field(None, alias="sl_endpoint_host")
    sl_endpoint_name: str | None = Field(None, alias="sl_endpoint_name")
    sl_endpoint_port: int | None = Field(None, alias="sl_endpoint_port")
    sl_host_original: str | None = Field(None, alias="sl_host_original")
    sl_http_proxy: bool | None = Field(None, alias="sl_http_proxy")
    sl_location_id: str | None = Field(None, alias="sl_location_id")
    sl_service_url: str | None = Field(None, alias="sl_service_url")
    ssl_certificate: str | None = Field(None, alias="ssl_certificate")
    ssl_certificate_hostname: str | None = Field(None, alias="ssl_certificate_host")
    validate_ssl_certificate: bool | None = Field(None, alias="ssl_certificate_validation")
    username: str = Field(None, alias="username")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    ssl_certificate_file: str | None = Field(None, alias="ssl_certificate_file")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        include.add("ssl_certificate") if (not self.ssl_certificate_file) else exclude.add("ssl_certificate")
        (
            include.add("access_token")
            if (((not self.defer_credentials)) and (not self.username))
            else exclude.add("access_token")
        )
        (
            include.add("password")
            if (
                (
                    ((not self.defer_credentials))
                    and ((not self.access_token))
                    and ((not self.use_my_platform_login_credentials))
                )
                and (
                    ((not self.defer_credentials))
                    and (
                        (
                            (
                                self.authentication_type
                                and (
                                    (
                                        hasattr(self.authentication_type, "value")
                                        and self.authentication_type.value == "basic"
                                    )
                                    or (self.authentication_type == "basic")
                                )
                            )
                        )
                        or (
                            self.authentication_type
                            and (
                                (
                                    hasattr(self.authentication_type, "value")
                                    and self.authentication_type.value == "bearer"
                                )
                                or (self.authentication_type == "bearer")
                            )
                        )
                        or (
                            self.authentication_type
                            and (
                                (
                                    hasattr(self.authentication_type, "value")
                                    and self.authentication_type.value == "cam_credentials"
                                )
                                or (self.authentication_type == "cam_credentials")
                            )
                        )
                    )
                )
            )
            else exclude.add("password")
        )
        include.add("sl_connector_id") if (not self.sl_location_id) else exclude.add("sl_connector_id")
        (
            include.add("gateway_url")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_type
                    and (
                        (hasattr(self.authentication_type, "value") and self.authentication_type.value == "bearer")
                        or (self.authentication_type == "bearer")
                    )
                )
                and (
                    (
                        (
                            self.authentication_type
                            and (
                                (
                                    hasattr(self.authentication_type, "value")
                                    and self.authentication_type.value != "basic"
                                )
                                or (self.authentication_type != "basic")
                            )
                        )
                    )
                    and (
                        self.authentication_type
                        and (
                            (
                                hasattr(self.authentication_type, "value")
                                and self.authentication_type.value != "cam_credentials"
                            )
                            or (self.authentication_type != "cam_credentials")
                        )
                    )
                )
            )
            else exclude.add("gateway_url")
        )
        (
            include.add("namespace")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_type
                    and (
                        (
                            hasattr(self.authentication_type, "value")
                            and self.authentication_type.value == "cam_credentials"
                        )
                        or (self.authentication_type == "cam_credentials")
                    )
                )
                and (
                    (
                        (
                            self.authentication_type
                            and (
                                (
                                    hasattr(self.authentication_type, "value")
                                    and self.authentication_type.value != "basic"
                                )
                                or (self.authentication_type != "basic")
                            )
                        )
                    )
                    and (
                        self.authentication_type
                        and (
                            (hasattr(self.authentication_type, "value") and self.authentication_type.value != "bearer")
                            or (self.authentication_type != "bearer")
                        )
                    )
                )
            )
            else exclude.add("namespace")
        )
        (
            include.add("ssl_certificate_hostname")
            if (((self.ssl_certificate)) and (self.validate_ssl_certificate))
            else exclude.add("ssl_certificate_hostname")
        )
        include.add("sl_location_id") if (not self.sl_connector_id) else exclude.add("sl_location_id")
        (
            include.add("username")
            if (
                (
                    ((not self.defer_credentials))
                    and ((not self.access_token))
                    and ((not self.use_my_platform_login_credentials))
                )
                and (
                    ((not self.defer_credentials))
                    and (
                        (
                            (
                                self.authentication_type
                                and (
                                    (
                                        hasattr(self.authentication_type, "value")
                                        and self.authentication_type.value == "basic"
                                    )
                                    or (self.authentication_type == "basic")
                                )
                            )
                        )
                        or (
                            self.authentication_type
                            and (
                                (
                                    hasattr(self.authentication_type, "value")
                                    and self.authentication_type.value == "bearer"
                                )
                                or (self.authentication_type == "bearer")
                            )
                        )
                        or (
                            self.authentication_type
                            and (
                                (
                                    hasattr(self.authentication_type, "value")
                                    and self.authentication_type.value == "cam_credentials"
                                )
                                or (self.authentication_type == "cam_credentials")
                            )
                        )
                    )
                )
            )
            else exclude.add("username")
        )
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )
        include.add("access_token") if (self.hidden_dummy_property1) else exclude.add("access_token")
        include.add("vaulted_properties") if (self.hidden_dummy_property1) else exclude.add("vaulted_properties")
        (
            include.add("use_my_platform_login_credentials")
            if (self.hidden_dummy_property1)
            else exclude.add("use_my_platform_login_credentials")
        )
        include.add("ssl_certificate_file") if (self.hidden_dummy_property1) else exclude.add("ssl_certificate_file")

        include.add("sl_connector_id") if (not self.sl_location_id) else exclude.add("sl_connector_id")
        (
            include.add("ssl_certificate_hostname")
            if (self.ssl_certificate) and (self.validate_ssl_certificate == "true" or self.validate_ssl_certificate)
            else exclude.add("ssl_certificate_hostname")
        )
        (
            include.add("namespace")
            if (not self.defer_credentials)
            and (
                self.authentication_type
                and (
                    (hasattr(self.authentication_type, "value") and self.authentication_type.value == "cam_credentials")
                    or (self.authentication_type == "cam_credentials")
                )
            )
            and (
                self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "basic" not in str(self.authentication_type.value)
                    )
                    or ("basic" not in str(self.authentication_type))
                )
                and self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "bearer" not in str(self.authentication_type.value)
                    )
                    or ("bearer" not in str(self.authentication_type))
                )
            )
            else exclude.add("namespace")
        )
        (
            include.add("gateway_url")
            if (not self.defer_credentials)
            and (
                self.authentication_type
                and (
                    (hasattr(self.authentication_type, "value") and self.authentication_type.value == "bearer")
                    or (self.authentication_type == "bearer")
                )
            )
            and (
                self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "basic" not in str(self.authentication_type.value)
                    )
                    or ("basic" not in str(self.authentication_type))
                )
                and self.authentication_type
                and (
                    (
                        hasattr(self.authentication_type, "value")
                        and self.authentication_type.value
                        and "cam_credentials" not in str(self.authentication_type.value)
                    )
                    or ("cam_credentials" not in str(self.authentication_type))
                )
            )
            else exclude.add("gateway_url")
        )
        include.add("ssl_certificate_file") if (not self.ssl_certificate) else exclude.add("ssl_certificate_file")
        (
            include.add("access_token")
            if (not self.defer_credentials) and (not self.username)
            else exclude.add("access_token")
        )
        (
            include.add("password")
            if (
                (not self.defer_credentials)
                and (not self.access_token)
                and (self.use_my_platform_login_credentials != "yes")
            )
            and (
                (not self.defer_credentials)
                and (
                    self.authentication_type
                    and (
                        (
                            hasattr(self.authentication_type, "value")
                            and self.authentication_type.value
                            and "basic" in str(self.authentication_type.value)
                        )
                        or ("basic" in str(self.authentication_type))
                    )
                    and self.authentication_type
                    and (
                        (
                            hasattr(self.authentication_type, "value")
                            and self.authentication_type.value
                            and "bearer" in str(self.authentication_type.value)
                        )
                        or ("bearer" in str(self.authentication_type))
                    )
                    and self.authentication_type
                    and (
                        (
                            hasattr(self.authentication_type, "value")
                            and self.authentication_type.value
                            and "cam_credentials" in str(self.authentication_type.value)
                        )
                        or ("cam_credentials" in str(self.authentication_type))
                    )
                )
            )
            else exclude.add("password")
        )
        (
            include.add("username")
            if (
                (not self.defer_credentials)
                and (not self.access_token)
                and (self.use_my_platform_login_credentials != "yes")
            )
            and (
                (not self.defer_credentials)
                and (
                    self.authentication_type
                    and (
                        (
                            hasattr(self.authentication_type, "value")
                            and self.authentication_type.value
                            and "basic" in str(self.authentication_type.value)
                        )
                        or ("basic" in str(self.authentication_type))
                    )
                    and self.authentication_type
                    and (
                        (
                            hasattr(self.authentication_type, "value")
                            and self.authentication_type.value
                            and "bearer" in str(self.authentication_type.value)
                        )
                        or ("bearer" in str(self.authentication_type))
                    )
                    and self.authentication_type
                    and (
                        (
                            hasattr(self.authentication_type, "value")
                            and self.authentication_type.value
                            and "cam_credentials" in str(self.authentication_type.value)
                        )
                        or ("cam_credentials" in str(self.authentication_type))
                    )
                )
            )
            else exclude.add("username")
        )
        include.add("ssl_certificate") if (not self.ssl_certificate_file) else exclude.add("ssl_certificate")
        include.add("sl_location_id") if (not self.sl_connector_id) else exclude.add("sl_location_id")
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



from ibm_watsonx_data_integration.services.datastage.models.flow.datastage_flow import DataStageFlow
from ibm_watsonx_data_integration.services.datastage.tests.integration.util import AbstractFlowRunTest
from os import environ
from ibm_watsonx_data_integration.cpd_models.project_model import Project


class TestSybaseConnection(AbstractFlowRunTest):
    def get_flow_name(self):
        return "sybase_peek"

    def get_fc(self, project: Project):
        fc = DataStageFlow(project=project, name=self.get_flow_name(), description="", flow_type="datastage")

        # Stages
        _testtable_1_ = fc.add_stage("SAP ASE", "testtable_1")
        _testtable_1_.configuration.schema_name = "tm_ds"
        _testtable_1_.configuration.table_name = "testtable"

        _testtable_1_.configuration.connection.database = "CONOPS"
        _testtable_1_.configuration.connection.hostname_or_ip_address = environ["SYBASE_HOST"]
        _testtable_1_.configuration.connection.password = environ["SYBASE_PASSWORD"]
        _testtable_1_.configuration.connection.port = "5000"
        _testtable_1_.configuration.connection.port_is_ssl_enabled = False
        _testtable_1_.configuration.connection.username = environ["SYBASE_USERNAME"]

        _peek_1_ = fc.add_stage("Peek", "Peek_1")
        _peek_1_.configuration.runtime_column_propagation = 0

        # Graph
        testtable_1_schema = _testtable_1_.connect_output_to(_peek_1_).set_name("Link_1").create_schema()

        # Schemas
        testtable_1_schema.add_field("INTEGER", "C1").nullable()
        testtable_1_schema.add_field("VARCHAR", "C2").nullable().length(20)

        project.update_flow(fc)

        return fc

    def validate(self, state):
        assert "Completed" in state



import collections
from ibm.datastage._framework.dag import (
    DAG,
    CommentNode,
    GhostNode,
    Link,
    Node,
)
from math import floor

Layer = list[Node]


class LayeredLayout:
    """Layered layout produces a graph layout by organizing nodes into layers ideal for visualizing data pipelines and
    other directed workflows. Earlier layers contain nodes with no incoming edges, while later layers contain nodes with
    no outgoing edges.

    The layout algorithm builds on top of the Sugiyama layered graph drawing technique.
    It consists of the following steps:
        1. Run DAG.compute_metadata() to assign useful node metadata such as x, y positioning, topological order,
           parents, children, etc.
        2. Collect comment nodes from the DAG and set their dimensions.
        3. Insert ghost (fake) nodes for edges that span non-adjacent layers. These are used to reduce edge crossings
           caused by long edges that span multiple layers. The ghost nodes are removed after the layout is computed.
        4. Collect nodes into layers based on their topological order in the DAG so that nodes always have inputs from
           earlier layers and outputs to later layers.
        5. Order nodes within each layer based on the barycentric heuristic, which reduces edge crossings by ordering
           nodes based on the average order of their neighbors.
        6. Position nodes based on their layer number and order within the layer as an initial starting point.
        7. Iteratively refine node positions based on the average Y position of their neighbors to reduce edge crossings.
        8. Remove ghost nodes and round node positions to the nearest integer.
    """

    # Default node dimensions
    NODE_WIDTH = 48
    NODE_HEIGHT = 48
    COMMENT_WIDTH = 175
    COMMENT_HEIGHT = 60

    # Constraints for spacing
    MIN_NODE_SPACING = 48
    NODE_SPACING = 200
    LAYER_SPACING = 200
    COMMENT_LAYER_SPACING = 300

    # Estimated center of the canvas
    CANVAS_OFFSET_X = 800
    CANVAS_OFFSET_Y = 400

    # Number of iterations of node-swapping to minimize edge crossings. One iteration consists of two sweeps
    BARYCENTRIC_SWEEP_ITERATIONS = 2

    # Alpha value (0-1) for node position refinement. Higher values increase convergence speed but may cause oscillation
    ALPHA = 0.4
    # Number of iterations to refine node positions
    REFINEMENT_ITERATIONS = 10

    def __init__(self, dag: DAG):
        self.dag = dag
        self.layers: list[Layer] = []
        self.comments: list[CommentNode] = []

    def compute(self):
        """Compute the layered layout of the DAG, populating the x, y, and height (if non-default) properties of each node."""
        self.__collect_comment_nodes()
        self.__insert_ghost_nodes()
        self.__collect_layers()
        self.__order_layer_nodes()
        self.__position_nodes()

        for _ in range(self.REFINEMENT_ITERATIONS):
            self.__refine_node_positions()

        self.__remove_ghost_nodes()
        self.__round_node_positions()

        for layer in self.layers:
            for i, node in enumerate(layer):
                node.metadata["index"] = i

    def __collect_comment_nodes(self):
        """Collect comment nodes from the DAG and set their dimensions."""
        for node in self.dag.nodes():
            if not isinstance(node, CommentNode):
                continue

            node.metadata["width"] = self.COMMENT_WIDTH
            node.metadata["height"] = self.COMMENT_HEIGHT
            self.comments.append(node)

    def __insert_ghost_nodes(self):
        """Insert ghost nodes for edges that span non-adjacent layers. This helps reduce edge crossings caused by
        long edges that span multiple layers.

        For each edge from a parent to a child where child's layer > parent's layer + 1, create ghost nodes for each
        intermediate layer. Each ghost node is marked with metadata['ghost'] = True.
        """
        for parent in list(self.dag.nodes()):
            # Not necessary for comment nodes, as we will position these later
            if isinstance(parent, CommentNode):
                continue

            for child in list(parent.metadata["children"]):
                parent_layer_num = parent.metadata["layer"]
                child_layer_num = child.metadata["layer"]

                # Check if the edge spans more than one layer
                if child_layer_num > parent_layer_num + 1:
                    last: Node = parent
                    # Insert ghost nodes for each intermediate layer
                    for layer_num in range(parent_layer_num + 1, child_layer_num):
                        ghost_node = GhostNode()
                        ghost_node.metadata["layer"] = layer_num
                        ghost_node.metadata["children"] = []
                        ghost_node.metadata["parents"] = [last]
                        # Connect the previous node to the ghost node
                        self.dag.adj.setdefault(last, {})[ghost_node] = None
                        last.metadata["children"].append(ghost_node)
                        last = ghost_node

                    # Connect the last ghost node to the child node
                    last.metadata["children"].append(child)
                    child.metadata["parents"].append(last)
                    self.dag.adj.setdefault(last, {})[child] = None

    def __remove_ghost_nodes(self):
        """Remove previously inserted ghost nodes from the DAG."""
        clean_adj: dict[Node, dict[Node, Link]] = {}

        for from_node in self.dag.nodes():
            if from_node.metadata.get("ghost", False):
                continue

            # Add non-ghost parent nodes to the new adjacency list
            # If the parent has any non-ghost children, their links will also be added
            clean_adj.setdefault(from_node, {})
            for to_node, link in self.dag.adj[from_node].items():
                if not to_node.metadata.get("ghost", False):
                    # Only keep links where both the parent and child are not ghost nodes
                    clean_adj[from_node][to_node] = link

        self.dag.adj = clean_adj

    def __round_node_positions(self):
        """Round node positions to the nearest integer."""
        for node in self.dag.nodes():
            node.metadata["x"] = round(node.metadata["x"])
            node.metadata["y"] = round(node.metadata["y"])

    def __refine_node_positions(self):
        """Refine node positions based on the average Y position of their neighbors."""
        new_positions: dict[Node, float] = {}

        # Shift nodes vertically based on the average Y position of their neighbors
        for node in self.dag.nodes():
            # Fixed weights adjusting relative influence or "pull" of parent and child nodes
            parent_weight = 0.5
            child_weight = 0.5

            neighbor_y_sum = 0
            neighbor_y_weight = 0

            # Factor in the Y positions of parent nodes
            for parent in node.metadata["parents"]:
                neighbor_y_sum += (
                    parent.metadata["y"] + parent.metadata.get("height", self.NODE_HEIGHT) / 2
                ) * parent_weight
                neighbor_y_weight += parent_weight

            # Factor in the Y positions of child nodes
            for child in node.metadata["children"]:
                neighbor_y_sum += (
                    child.metadata["y"] + child.metadata.get("height", self.NODE_HEIGHT) / 2
                ) * child_weight
                neighbor_y_weight += child_weight

            if neighbor_y_weight:
                # Compute target Y position based on average Y position of neighbors
                target_y = neighbor_y_sum / neighbor_y_weight - node.metadata.get("height", self.NODE_HEIGHT) / 2
                # Move node towards (but not fully onto) target Y position based on ALPHA
                new_y = node.metadata["y"] + self.ALPHA * (target_y - node.metadata["y"])
                new_positions[node] = new_y

        # Replace old Y positions with new Y positions
        for node, new_y in new_positions.items():
            node.metadata["y"] = new_y

        # Prevent nodes from overlapping by ensuring a minimum vertical spacing between sibling nodes
        for layer in self.layers:
            # Skip empty layers
            if not layer:
                continue

            for i in range(1, len(layer)):
                prev_node = layer[i - 1]
                cur_node = layer[i]
                # Compute desired vertical space between sibling nodes
                height_space = prev_node.metadata.get("height", self.NODE_HEIGHT)
                desired_space = self.MIN_NODE_SPACING + height_space
                gap = cur_node.metadata["y"] - prev_node.metadata["y"]
                if gap < desired_space:
                    # If nodes are too close, move the current node down (higher Y)
                    cur_node.metadata["y"] += desired_space - gap

            # Center the entire layer vertically on the canvas
            all_y = [node.metadata["y"] + node.metadata.get("height", self.NODE_HEIGHT) / 2 for node in layer]
            center_y = (min(all_y) + max(all_y)) / 2
            for node in layer:
                node.metadata["y"] += self.CANVAS_OFFSET_Y - center_y

    @staticmethod
    def __count_node_relatives(node: Node):
        """Count the number of relatives (parents and children) of a node, excluding ghost nodes."""

        def is_relative(n: Node):
            return not n.metadata.get("ghost", False) and not isinstance(n, CommentNode)

        return len([n for n in node.metadata["parents"] + node.metadata["children"] if is_relative(n)])

    def __position_nodes(self):
        """Position nodes based on their layer number and order within the layer as an initial starting point.
        Also adjust the height of nodes with many relatives to improve visual distinction.
        """
        x_pos = 0
        max_x_pos = 0
        max_y_pos = 0

        for layer_num, layer in enumerate(self.layers):
            y_pos = 0
            has_comment = False

            for node_num, node in enumerate(layer):
                # Resize non-comment nodes if conditions are met
                if not isinstance(node, CommentNode):
                    if (relatives := self.__count_node_relatives(node)) >= 3:
                        # Increase height of nodes with many relatives (either many parents or many children)
                        node.metadata["height"] = (self.NODE_HEIGHT + self.MIN_NODE_SPACING / 2) * relatives
                else:
                    has_comment = True

                node.metadata["x"] = x_pos
                node.metadata["y"] = y_pos
                y_pos += self.NODE_SPACING
                max_x_pos = max(max_x_pos, x_pos)
                max_y_pos = max(max_y_pos, y_pos)

            if has_comment:
                x_pos += self.COMMENT_LAYER_SPACING
            else:
                x_pos += self.LAYER_SPACING

        # Center entire flow
        x_offset = x_pos / 2
        y_offset = max_y_pos / 2

        for node in self.dag.nodes():
            node.metadata["x"] += self.CANVAS_OFFSET_X - x_offset
            node.metadata["y"] += self.CANVAS_OFFSET_Y - y_offset

    def __order_layer_nodes(self):
        """Order nodes within each layer based on the barycentric heuristic.

        The barycentric heuristic reduces edge crossings by ordering nodes in one layer based on the average order of
        their neighbors. We perform alternating downward and upward sweeps to iteratively improve the ordering of nodes
        within each layer.
        """
        # Initially order nodes within each layer alphabetically for consistency
        for layer in self.layers:
            layer.sort(key=lambda n: type(n).__name__)

        # Then order nodes based on barycentric heuristic
        for _ in range(self.BARYCENTRIC_SWEEP_ITERATIONS):
            # Downward sweep - order layers from top to bottom
            for i in range(1, len(self.layers)):
                barycenters = {}
                for node in self.layers[i]:
                    barycenters[node] = self.__compute_weighted_barycenter(node, i)
                self.layers[i].sort(key=lambda n: barycenters[n])

            # Upward sweep - order layers from bottom to top
            for i in range(len(self.layers) - 2, -1, -1):
                barycenters = {}
                for node in self.layers[i]:
                    barycenters[node] = self.__compute_weighted_barycenter(node, i)
                self.layers[i].sort(key=lambda n: barycenters[n])

    def __compute_normalized_index(self, node: Node, current_layer_num: int):
        """Computes the continuous normalized index of a node within its layer.
        0.5 means the node is in the middle of the layer.
        0 means the node is the first in the layer.
        1 means the node is last.
        """
        layer = self.layers[current_layer_num]
        if len(layer) == 1:
            return 0.5
        return layer.index(node) / (len(layer) - 1)

    def __compute_weighted_barycenter(self, node: Node, current_layer_num: int):
        """Computes the barycenter of a node within its layer. The computation discounts the influence of neighbor nodes
        that are in layers further away from the current layer.
        """
        total_weight = 0
        weighted_sum = 0

        for parent in node.metadata["parents"]:
            parent_layer_num = parent.metadata["layer"]
            if parent_layer_num < current_layer_num:
                weight = 1 / (current_layer_num - parent_layer_num)
                weighted_sum += weight * self.__compute_normalized_index(parent, parent_layer_num)
                total_weight += weight

        for child in node.metadata["children"]:
            child_layer_num = child.metadata["layer"]
            if child_layer_num > current_layer_num:
                weight = 1 / (child_layer_num - current_layer_num)
                weighted_sum += weight * self.__compute_normalized_index(child, child_layer_num)
                total_weight += weight

        if total_weight > 0:
            return weighted_sum / total_weight
        else:
            return 0.5

    def __collect_layers(self):
        """Collect nodes into list of layers based on the topological order of the DAG."""
        assert self.dag.adj, "DAG is empty"

        # Collect nodes into a dictionary mapping the layer # to the Layer
        layer_dict: dict[int, Layer] = collections.defaultdict(list)
        max_layer = 0
        for node in self.dag.nodes():
            # Don't include comment nodes in layers yet
            if isinstance(node, CommentNode):
                continue
            layer_num = node.metadata["layer"]
            max_layer = max(max_layer, layer_num)
            layer_dict[layer_num].append(node)

        # Assign comment nodes to layers
        for comment in self.comments:
            children = comment.metadata["children"]
            if not children:
                layer_dict[0].append(comment)
                continue

            layer_num = floor(sum([child.metadata["layer"] for child in children]) / len(children))
            comment.metadata["layer"] = layer_num
            layer_dict[layer_num].append(comment)

        # Convert the dictionary mapping to an ordered list of Layers
        for layer_num in range(max_layer + 1):
            self.layers.append(layer_dict[layer_num])



import warnings
from ibm_watsonx_data_integration.services.datastage.models.base_stage import BaseStage
from ibm_watsonx_data_integration.services.datastage.models.enums import INVESTIGATE
from pydantic import Field
from typing import ClassVar


class investigate(BaseStage):

    op_name: ClassVar[str] = "Investigate"
    node_type: ClassVar[str] = "execution_node"
    image: ClassVar[str] = "/data-intg/flows/graphics/palette/Investigate.svg"
    label: ClassVar[str] = "Investigate"
    runtime_column_propagation: bool = Field(False, alias="runtime_column_propagation")
    alpha: bool | None = Field(False, alias="alpha")
    buf_free_run: int | None = Field(50, alias="buf_free_run")
    buf_free_run_ronly: str | None = Field("", alias="buf_free_run_ronly")
    buf_mode: INVESTIGATE.BufMode | None = Field(INVESTIGATE.BufMode.default, alias="buf_mode")
    buf_mode_ronly: str | None = Field("", alias="buf_mode_ronly")
    char_concat: str | None = Field("", alias="charConcat")
    char_discrete: bool | None = Field(False, alias="charDiscrete")
    choose_char_concate: bool | None = Field(False, alias="chooseCharConcate")
    coll_type: INVESTIGATE.CollType | None = Field(INVESTIGATE.CollType.auto, alias="coll_type")
    columns_invt: list | None = Field([], alias="columnsInvt")
    combinability: INVESTIGATE.Combinability | None = Field(INVESTIGATE.Combinability.auto, alias="combinability")
    compress_unknown: bool | None = Field(False, alias="compressUnknown")
    db2_instance_select: INVESTIGATE.Db2InstanceSelect | None = Field(
        INVESTIGATE.Db2InstanceSelect.use_db2InstanceEnv, alias="db2InstanceSelect"
    )
    db2_name_select: INVESTIGATE.Db2NameSelect | None = Field(
        INVESTIGATE.Db2NameSelect.use_db2NameEnv, alias="db2NameSelect"
    )
    disk_write_inc: int | None = Field(1048576, alias="disk_write_inc")
    disk_write_inc_ronly: str | None = Field("", alias="disk_write_inc_ronly")
    enable_schemaless_design: bool = Field(False, alias="enableSchemalessDesign")
    execmode: INVESTIGATE.Execmode | None = Field(INVESTIGATE.Execmode.default_par, alias="execmode")
    field: list = Field([], alias="field")
    flow_dirty: str | None = Field("false", alias="flow_dirty")
    hide: bool | None = Field(False, alias="hide")
    input_count: int | None = Field(0, alias="input_count")
    input_link_description: list | None = Field("", alias="inputLinkDescription")
    inputcol_properties: list | None = Field([], alias="inputcolProperties")
    key_col_select: INVESTIGATE.KeyColSelect | None = Field(
        INVESTIGATE.KeyColSelect.Select_a_column, alias="keyColSelect"
    )
    key_cols_coll: list | None = Field([], alias="keyColsColl")
    key_cols_coll_ordered: list | None = Field([], alias="keyColsCollOrdered")
    key_cols_coll_robin: list | None = Field([], alias="keyColsCollRobin")
    key_cols_none: list | None = Field([], alias="keyColsNone")
    key_cols_part: list | None = Field([], alias="keyColsPart")
    locale: str | None = Field(None, alias="locale")
    mask_properties: list | None = Field([], alias="maskProperties")
    max_mem_buf_size: int | None = Field(3145728, alias="max_mem_buf_size")
    max_mem_buf_size_ronly: str | None = Field("", alias="max_mem_buf_size_ronly")
    min_freq: int = Field(0, alias="minFreq")
    numeric: bool | None = Field(False, alias="numeric")
    output_count: int | None = Field(0, alias="output_count")
    output_link_description: list | None = Field("", alias="outputLinkDescription")
    outputcol_properties: list | None = Field([], alias="outputcolProperties")
    outputlink_ordering_list: list | None = Field([], alias="OutputlinkOrderingList")
    part_dbname: str | None = Field("eg: SAMPLE", alias="part_dbname")
    part_server: str | None = Field("eg: DB2INST", alias="part_server")
    part_stable: bool | None = Field(None, alias="part_stable")
    part_stable_coll: bool | None = Field(False, alias="part_stable_coll")
    part_table: str | None = Field(None, alias="part_table")
    part_type: INVESTIGATE.PartType | None = Field(INVESTIGATE.PartType.auto, alias="part_type")
    part_unique: bool | None = Field(None, alias="part_unique")
    part_unique_coll: bool | None = Field(False, alias="part_unique_coll")
    pattern: list | None = Field([], alias="pattern")
    pattern_rpt: bool | None = Field(False, alias="patternRpt")
    perform_sort: bool | None = Field(False, alias="perform_sort")
    perform_sort_coll: bool | None = Field(False, alias="perform_sort_coll")
    perform_sort_modulus: bool | None = Field(False, alias="perform_sort_modulus")
    preserve: INVESTIGATE.Preserve | None = Field(INVESTIGATE.Preserve.default_propagate, alias="preserve")
    pttn_rpt: int | None = Field(0, alias="pttnRpt")
    q_s_x_m_l: str = Field(None, alias="QSXML")
    queue_upper_size: int | None = Field(0, alias="queue_upper_size")
    queue_upper_size_ronly: str | None = Field("", alias="queue_upper_size_ronly")
    ruleset: str | None = Field(None, alias="ruleset")
    runtime_column_propagation: bool | None = Field(None, alias="runtime_column_propagation")
    schema_name_check: bool | None = Field(None, alias="schemaNameCheck")
    seplist: str = Field("XX !?%$,.;:()-/#&", alias="seplist")
    show_coll_type: int | None = Field(0, alias="showCollType")
    show_part_type: int | None = Field(1, alias="showPartType")
    show_sort_options: int | None = Field(0, alias="showSortOptions")
    sort_instructions: str | None = Field(
        "If sort keys are added, a sort will be performed.", alias="sortInstructions"
    )
    special: bool | None = Field(False, alias="special")
    stage_description: list | None = Field("", alias="stageDescription")
    stasticsoptions: str | None = Field(None, alias="Stasticsoptions")
    striplist: str = Field("XX !?%$,.;:()-/#&", alias="striplist")
    token_edit: INVESTIGATE.TokenEdit | None = Field(INVESTIGATE.TokenEdit.abbrev, alias="tokenEdit")
    tokn_rpt: bool | None = Field(False, alias="toknRpt")
    trace_properties: list | None = Field([], alias="traceProperties")
    word: INVESTIGATE.Word | None = Field(INVESTIGATE.Word.true, alias="word")
    workdir: str = Field("", alias="workdir")

    def validate_parameters(self) -> tuple[set, set]:
        include = set()
        exclude = set()

        include.add("max_mem_buf_size") if (self.buf_mode != "nobuffer") else exclude.add("max_mem_buf_size")
        include.add("buf_free_run") if (self.buf_mode != "nobuffer") else exclude.add("buf_free_run")
        include.add("queue_upper_size") if (self.buf_mode != "nobuffer") else exclude.add("queue_upper_size")
        include.add("disk_write_inc") if (self.buf_mode != "nobuffer") else exclude.add("disk_write_inc")

        (
            include.add("runtime_column_propagation")
            if (not self.enable_schemaless_design)
            else exclude.add("runtime_column_propagation")
        )
        (
            include.add("max_mem_buf_size_ronly")
            if (self.hide != "true" or not self.hide)
            else exclude.add("max_mem_buf_size_ronly")
        )
        (
            include.add("buf_free_run_ronly")
            if (self.hide != "true" or not self.hide)
            else exclude.add("buf_free_run_ronly")
        )
        (
            include.add("queue_upper_size_ronly")
            if (self.hide != "true" or not self.hide)
            else exclude.add("queue_upper_size_ronly")
        )
        (
            include.add("disk_write_inc_ronly")
            if (self.hide != "true" or not self.hide)
            else exclude.add("disk_write_inc_ronly")
        )
        include.add("mask_properties") if (self.word and " " in str(self.word)) else exclude.add("mask_properties")
        include.add("columns_invt") if (self.word and "word" in str(self.word)) else exclude.add("columns_invt")
        include.add("striplist") if (self.word and "word" in str(self.word)) else exclude.add("striplist")
        include.add("seplist") if (self.word and "word" in str(self.word)) else exclude.add("seplist")
        include.add("ruleset") if (()) else exclude.add("ruleset")
        (
            include.add("compress_unknown")
            if ((()) and (self.word and "word" in str(self.word)))
            else exclude.add("compress_unknown")
        )
        include.add("numeric") if ((()) and (self.word and "word" in str(self.word))) else exclude.add("numeric")
        include.add("alpha") if ((()) and (self.word and "word" in str(self.word))) else exclude.add("alpha")
        include.add("special") if ((()) and (self.word and "word" in str(self.word))) else exclude.add("special")
        include.add("pattern_rpt") if (self.word and "word" in str(self.word)) else exclude.add("pattern_rpt")
        include.add("tokn_rpt") if (self.word and "word" in str(self.word)) else exclude.add("tokn_rpt")
        (
            include.add("choose_char_concate")
            if (self.word and " " in str(self.word))
            else exclude.add("choose_char_concate")
        )
        (
            include.add("part_stable")
            if (((self.show_part_type)) and (self.part_type != "auto") and (self.show_sort_options))
            else exclude.add("part_stable")
        )
        (
            include.add("part_unique")
            if (((self.show_part_type)) and (self.part_type != "auto") and (self.show_sort_options))
            else exclude.add("part_unique")
        )
        (
            include.add("key_cols_part")
            if (
                (((self.show_part_type)) and ((self.part_type != "auto")) and ((self.part_type != "modulus")))
                or (
                    ((self.show_part_type))
                    and (self.part_type == "modulus")
                    and (self.perform_sort_modulus)
                )
            )
            else exclude.add("key_cols_part")
        )
        (
            include.add("db2_name_select")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("db2_name_select")
        )
        (
            include.add("part_dbname")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_dbname")
        )
        (
            include.add("db2_instance_select")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("db2_instance_select")
        )
        (
            include.add("part_server")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_server")
        )
        (
            include.add("part_table")
            if (((self.part_type == "db2part")) and (self.show_part_type))
            else exclude.add("part_table")
        )
        (
            include.add("perform_sort")
            if (((self.part_type == "hash")) or (self.part_type == "range"))
            else exclude.add("perform_sort")
        )
        include.add("perform_sort_modulus") if (self.part_type == "modulus") else exclude.add("perform_sort_modulus")
        (
            include.add("key_col_select")
            if (((self.part_type == "modulus")) and (not self.perform_sort_modulus))
            else exclude.add("key_col_select")
        )
        (
            include.add("sort_instructions")
            if (
                ((self.show_part_type))
                and (
                    ((self.part_type == "db2part"))
                    or (self.part_type == "entire")
                    or (self.part_type == "random")
                    or (self.part_type == "roundrobin")
                    or (self.part_type == "same")
                )
            )
            else exclude.add("sort_instructions")
        )
        include.add("coll_type") if (self.show_coll_type) else exclude.add("coll_type")
        include.add("part_type") if (self.show_part_type) else exclude.add("part_type")
        (
            include.add("perform_sort_coll")
            if (((self.coll_type == "ordered")) or (self.coll_type == "roundrobin_coll"))
            else exclude.add("perform_sort_coll")
        )
        (
            include.add("key_cols_coll")
            if (
                ((self.show_coll_type))
                and (((self.coll_type == "sortmerge")) or (self.perform_sort_coll))
            )
            else exclude.add("key_cols_coll")
        )
        (
            include.add("key_cols_none")
            if (((not self.show_part_type)) and (not self.show_coll_type))
            else exclude.add("key_cols_none")
        )
        (
            include.add("part_stable_coll")
            if (
                ((self.coll_type == "sortmerge"))
                or (self.perform_sort_coll)
                or (((not self.show_part_type)) and (not self.show_coll_type))
            )
            else exclude.add("part_stable_coll")
        )
        (
            include.add("part_unique_coll")
            if (
                ((self.coll_type == "sortmerge"))
                or (self.perform_sort_coll)
                or (((not self.show_part_type)) and (not self.show_coll_type))
            )
            else exclude.add("part_unique_coll")
        )
        return include, exclude

    def get_input_ports_props(self) -> dict:
        include, exclude = self.validate()
        props = {"runtime_column_propagation"}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_input_cardinality(self) -> dict:
        return {"min": 1, "max": 1}

    def get_output_ports_props(self) -> dict:
        include, exclude = self.validate()
        props = set()
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_output_cardinality(self) -> dict:
        return {"min": 1, "max": 2}

    def get_parameters_props(self) -> dict:
        include, exclude = self.validate()
        props = {
            "additional_properties_set_options",
            "additional_properties_set_token_options",
            "alpha",
            "buf_free_run",
            "buf_free_run_ronly",
            "buf_mode",
            "buf_mode_ronly",
            "char_concat",
            "char_discrete",
            "choose_char_concate",
            "coll_type",
            "columns_invt",
            "combinability",
            "compress_unknown",
            "db2_instance_select",
            "db2_name_select",
            "disk_write_inc",
            "disk_write_inc_ronly",
            "enable_schemaless_design",
            "execmode",
            "field",
            "flow_dirty",
            "hide",
            "input_count",
            "input_link_description",
            "inputcol_properties",
            "key_col_select",
            "key_cols_coll",
            "key_cols_coll_ordered",
            "key_cols_coll_robin",
            "key_cols_none",
            "key_cols_part",
            "locale",
            "mask_properties",
            "max_mem_buf_size",
            "max_mem_buf_size_ronly",
            "min_freq",
            "numeric",
            "output_count",
            "output_link_description",
            "outputcol_properties",
            "outputlink_ordering_list",
            "part_dbname",
            "part_server",
            "part_stable",
            "part_stable_coll",
            "part_table",
            "part_type",
            "part_unique",
            "part_unique_coll",
            "pattern",
            "pattern_rpt",
            "perform_sort",
            "perform_sort_coll",
            "perform_sort_modulus",
            "preserve",
            "pttn_rpt",
            "q_s_x_m_l",
            "queue_upper_size",
            "queue_upper_size_ronly",
            "ruleset",
            "runtime_column_propagation",
            "schema_name_check",
            "seplist",
            "show_coll_type",
            "show_part_type",
            "show_sort_options",
            "sort_instructions",
            "special",
            "stage_description",
            "stasticsoptions",
            "striplist",
            "token_edit",
            "tokn_rpt",
            "trace_properties",
            "word",
            "workdir",
        }
        required = {
            "enable_schemaless_design",
            "field",
            "min_freq",
            "pattern",
            "q_s_x_m_l",
            "ruleset",
            "word",
            "workdir",
        }
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_app_data_props(self) -> dict:
        return {
            "datastage": {
                "maxRejectOutputs": 0,
                "minRejectOutputs": 0,
                "maxReferenceInputs": 0,
                "minReferenceInputs": 0,
            }
        }



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""This file contains an empty pipeline definition, and a method to create a pipeline definition for new flows."""

from copy import deepcopy
from uuid import uuid4

EMPTY_PIPELINE_DEFINITION = {
    "schemaVersion": 6,
    "version": 50,
    "pipelineId": "use a flow's executor id here",
    "title": "use a flow's executor id here",
    "description": "description",  # ignore since it's only in the datacollector
    "uuid": "create a new uuid here",
    "configuration": [
        {"name": "transformerEmrConnection.retryPolicyConfig.throttledBaseDelay", "value": 500},
        {"name": "googleCloudCredentialsConfig.credentialsProvider", "value": "DEFAULT_PROVIDER"},
        {"name": "googleCloudConfig.customRegion", "value": None},
        {"name": "googleCloudConfig.workerCount", "value": 2},
        {"name": "udfSchema", "value": None},
        {"name": "transformerEmrConnection.defineBootstrapActions", "value": False},
        {"name": "emrServerlessConnection.stopApplication", "value": True},
        {"name": "udfDb", "value": None},
        {"name": "transformerEmrConnection.awsConfig.credentialMode", "value": "WITH_CREDENTIALS"},
        {"name": "livyConfig.baseUrlHD", "value": "https://localhost/livy/"},
        {"name": "useSnowflakeConnection", "value": "false"},
        {"name": "emrServerlessConnection.retryPolicyConfig.throttledBaseDelay", "value": 500},
        {"name": "transformerEmrConnection.provisionedProductName", "value": None},
        {"name": "googleCloudCredentialsConfig.path", "value": None},
        {"name": "shouldRetry", "value": True},
        {"name": "transformerEmrConnection.clusterId", "value": None},
        {"name": "emrServerlessConnection.emrApplicationTags", "value": []},
        {"name": "transformerEmrConnection.productId", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.oauthClientSecret", "value": None},
        {"name": "transformerEmrConnection.awsConfig.sessionDuration", "value": 3600},
        {"name": "clusterConfig.yarnKerberosKeytabBase64Bytes", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.customUrl", "value": "jdbc:snowflake://[HOST]:[PORT]"},
        {"name": "emrServerlessConnection.applicationByNameAndTags", "value": False},
        {"name": "snowflakeCommonConfig.overrideRole", "value": None},
        {"name": "runnerIdleTIme", "value": 60},
        {"name": "emrServerlessConnectionSelection", "value": "MANUAL"},
        {"name": "executionMode", "value": "STANDALONE"},
        {"name": "snowflakeCommonConfig.snowflakeConnection.usePrivateLinkURL", "value": False},
        {"name": "snowflakeCommonConfig.snowflakeConnection.account", "value": None},
        {"name": "snowflakeCommonConfig.overrideWarehouse", "value": None},
        {"name": "warehouse", "value": None},
        {"name": "emrServerlessConnection.awsConfig.roleARN", "value": "arn:aws:iam::<account-id>:role/role-name"},
        {"name": "errorInformationLevel", "value": "FULL_ERROR_INFORMATION"},
        {"name": "schema", "value": None},
        {"name": "role", "value": None},
        {"name": "clusterConfig.useYarnKerberosKeytab", "value": False},
        {"name": "googleCloudConfig.gcsStagingUri", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.snowflakeAuthenticationMethod", "value": "KEYPAIR_CONTENT"},
        {"name": "logLevel", "value": "INFO"},
        {"name": "googleCloudConfig.region", "value": None},
        {"name": "googleCloudConfig.network", "value": None},
        {"name": "retryAttempts", "value": -1},
        {"name": "snowflakeCommonConfig.connectionSelection", "value": "MANUAL"},
        {"name": "databricksConfig.enableExternalResourceForExecutors", "value": True},
        {"name": "transformerEmrConnection.customTags", "value": []},
        {"name": "emrServerlessConnection.awsConfig.profileName", "value": None},
        {"name": "errorRecordPolicy", "value": "ORIGINAL_RECORD"},
        {"name": "clusterConfig.deployMode", "value": "CLIENT"},
        {"name": "transformerEmrConnection.terminateCluster", "value": True},
        {"name": "skipMetrics", "value": False},
        {"name": "snowflakeCommonConfig.snowflakeConnection.database", "value": None},
        {"name": "transformerEmrConnection.generateProductName", "value": False},
        {"name": "snowflakeCommonConfig.snowflakeConnection.password", "value": None},
        {"name": "emrServerlessConnection.retryPolicyConfig.baseDelay", "value": 100},
        {"name": "livyConfig.password", "value": None},
        {"name": "transformerRetryAttempts", "value": 3},
        {"name": "transformerEmrConnection.clusterTags", "value": []},
        {"name": "db", "value": None},
        {"name": "transformerEmrConnection.s3LogUri", "value": None},
        {"name": "googleCloudConfig.clusterName", "value": None},
        {"name": "transformerEmrConnection.awsConfig.awsSTSRegionConfig.customSigningRegion", "value": None},
        {"name": "transformerEmrConnection.awsConfig.awsAccessKeyId", "value": None},
        {"name": "googleCloudConfig.subnet", "value": None},
        {"name": "notifyOnStates", "value": ["RUN_ERROR", "STOPPED", "FINISHED"]},
        {"name": "emrServerlessConnection.s3LogUri", "value": None},
        {"name": "livyConfig.username", "value": None},
        {"name": "transformerEmrConnection.awsConfig.setSessionTags", "value": True},
        {"name": "transformerEmrConnection.loggingEnabled", "value": True},
        {"name": "googleCloudCredentialsConfig.credentialsFileContent", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.schema", "value": None},
        {"name": "googleCloudConfig.workerType", "value": None},
        {"name": "transformerEmrConnection.clusterByNameAndTags", "value": False},
        {"name": "transformerEmrConnection.bootstrapActionScripts", "value": []},
        {"name": "cdeConfig.resourceFilePrefix", "value": "streamsets"},
        {"name": "transformerEmrConnection.clusterPrefix", "value": None},
        {
            "name": "testOriginStage",
            "value": "streamsets-datacollector-dev-lib::"
            "com_streamsets_pipeline_stage_devtest_rawdata_RawDataDSource::3",
        },
        {"name": "emrServerlessConnection.awsConfig.awsSTSRegionConfig.customSigningRegion", "value": None},
        {"name": "emrServerlessConnection.maximumMemory", "value": 3000},
        {"name": "transformerEmrConnection.region", "value": "US_WEST_2"},
        {"name": "clusterConfig.dataCacheReplicas", "value": 1},
        {"name": "transformerEmrConnectionSelection", "value": "MANUAL"},
        {"name": "transformerEmrConnection.awsConfig.awsSTSRegionConfig.signingRegion", "value": "US_WEST_2"},
        {"name": "transformerEmrConnection.serviceAccessSecurityGroup", "value": None},
        {"name": "ludicrousMode", "value": False},
        {"name": "databricksConfig.useUnityCatalog", "value": False},
        {"name": "emrServerlessConnection.awsConfig.awsSTSRegionConfig.regionalEndpoint", "value": None},
        {"name": "shouldCreateFailureSnapshot", "value": True},
        {"name": "databricksConfig.token", "value": None},
        {"name": "transformerEmrConnection.primaryInstanceType", "value": None},
        {"name": "emailIDs", "value": []},
        {"name": "transformerEmrConnection.ec2SubnetId", "value": None},
        {"name": "emrServerlessConnection.applicationPrefix", "value": None},
        {"name": "transformerEmrConnection.stepConcurrency", "value": 10},
        {"name": "cdeConfig.workloadPassword", "value": None},
        {
            "name": "stopEventStage",
            "value": "streamsets-datacollector-basic-lib::"
            "com_streamsets_pipeline_stage_destination_devnull_ToErrorNullDTarget::1",
        },
        {"name": "clusterConfig.hadoopUserName", "value": None},
        {"name": "edgeHttpUrl", "value": "http://localhost:18633"},
        {"name": "transformerEMRConfig.executionRole", "value": None},
        {"name": "transformerEmrConnection.visibleToAllUsers", "value": True},
        {"name": "transformerEmrConnection.s3StagingUri", "value": None},
        {"name": "databricksConfig.initScriptsContents", "value": []},
        {"name": "transformerEmrConnection.secondaryInstanceTypeCustom", "value": None},
        {"name": "emrServerlessConnection.emrApplicationName", "value": None},
        {"name": "transformerEmrConnection.clusterName", "value": None},
        {"name": "clusterConfig.sparkPrimaryUrl", "value": "local[*]"},
        {"name": "emrServerlessConnection.awsConfig.roleSessionName", "value": None},
        {"name": "transformerEmrConnection.bootstrapActionSource", "value": "IN_S3"},
        {"name": "snowflakeCommonConfig.snowflakeConnection.oauthClientId", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.privateKeyPassword", "value": None},
        {"name": "googleCloudConfig.networkType", "value": None},
        {"name": "transformerEmrConnection.secondarySecurityGroup", "value": None},
        {"name": "transformerEmrConnection.awsConfig.externalID", "value": None},
        {"name": "googleCloudConfig.primaryType", "value": None},
        {"name": "clusterConfig.yarnKerberosKeytabSource", "value": "PROPERTIES_FILE"},
        {"name": "triggerInterval", "value": 2000},
        {"name": "transformerEmrConnection.primarySecurityGroup", "value": None},
        {"name": "emrServerlessConnection.subnetIds", "value": []},
        {"name": "emrServerlessConnection.maximumCpu", "value": 400},
        {"name": "transformerEmrConnection.serviceRole", "value": "EMR_DefaultRole"},
        {
            "name": "databricksConfig.clusterConfig",
            "value": '{\n    "num_workers": 8,\n    "spark_version": "14.3.x-scala2.12",'
            '\n    "node_type_id": "i3.xlarge"\n}',
        },
        {"name": "googleCloudCredentialsConfig.projectId", "value": None},
        {"name": "livyConfig.baseUrlBDC", "value": "https://localhost:30443/gateway/default/livy/v1/"},
        {"name": "transformerEmrConnection.customRegion", "value": None},
        {"name": "transformerEmrConnection.secondaryInstanceType", "value": None},
        {"name": "cdeConfig.resource", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.privateKeyPemContent", "value": None},
        {"name": "cdeConfig.apiUrl", "value": None},
        {"name": "connectionString", "value": None},
        {"name": "databricksConfig.provisionNewCluster", "value": True},
        {
            "name": "statsAggregatorStage",
            "value": "streamsets-datacollector-basic-lib::"
            "com_streamsets_pipeline_stage_destination_devnull_StatsDpmDirectlyDTarget::1",
        },
        {"name": "udfConfigBeans", "value": []},
        {"name": "databricksConfig.password", "value": None},
        {"name": "transformerEmrConnection.jobFlowRole", "value": "EMR_EC2_DefaultRole"},
        {"name": "snowflakeCommonConfig.overrideSchema", "value": None},
        {"name": "emrServerlessConnection.awsConfig.awsSTSRegionConfig.nonRegionalEndpoint", "value": None},
        {"name": "clusterConfig.stagingDir", "value": "/streamsets"},
        {"name": "transformerEmrConnection.awsConfig.awsSTSRegionConfig.regionalEndpoint", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.user", "value": None},
        {"name": "emrServerlessConnection.region", "value": "US_WEST_2"},
        {"name": "cdeConfig.workloadUser", "value": None},
        {"name": "emrServerlessConnection.createNewApplication", "value": False},
        {"name": "databricksConfig.clusterId", "value": None},
        {"name": "deliveryGuarantee", "value": "AT_LEAST_ONCE"},
        {"name": "maxRunners", "value": 0},
        {"name": "databricksConfig.credentialType", "value": "TOKEN"},
        {"name": "emrServerlessConnection.awsConfig.credentialMode", "value": "WITH_CREDENTIALS"},
        {"name": "transformerEmrConnection.emrClusterOption", "value": "EXISTING_CLUSTER"},
        {"name": "transformerEmrConnection.retryPolicyConfig.maxErrorRetry", "value": 6},
        {"name": "transformerEmrConnection.terminateProvisionedProduct", "value": True},
        {"name": "rateLimit", "value": 0},
        {"name": "transformerEmrConnection.retryPolicyConfig.maxBackoff", "value": 20000},
        {"name": "transformerEmrConnection.instanceCount", "value": 2},
        {"name": "emrServerlessConnection.s3StagingUri", "value": None},
        {
            "name": "badRecordsHandling",
            "value": "streamsets-datacollector-basic-lib::"
            "com_streamsets_pipeline_stage_destination_devnull_ToErrorNullDTarget::1",
        },
        {"name": "transformerEmrConnection.awsConfig.awsSecretAccessKey", "value": None},
        {"name": "clusterConfig.yarnKerberosKeytab", "value": None},
        {"name": "databricksConfig.username", "value": None},
        {"name": "googleCloudConfig.tags", "value": []},
        {"name": "emrServerlessConnection.customTags", "value": []},
        {"name": "transformerEmrConnection.awsConfig.isAssumeRole", "value": False},
        {"name": "transformerEmrConnection.awsConfig.awsSTSRegionConfig.region", "value": "US_WEST_2"},
        {"name": "transformerEmrConnection.awsConfig.awsSTSRegionConfig.nonRegionalEndpoint", "value": None},
        {"name": "emrServerlessConnection.awsConfig.awsSTSRegionConfig.customRegion", "value": None},
        {"name": "googleCloudConfig.create", "value": False},
        {"name": "emrServerlessConnection.awsConfig.externalID", "value": None},
        {"name": "transformerEMRConfig.encryption", "value": "NONE"},
        {"name": "emrServerlessConnection.loggingEnabled", "value": True},
        {"name": "googleCloudConfig.version", "value": "1.4-ubuntu18"},
        {"name": "snowflakeCommonConfig.snowflakeConnection.snowflakeRegion", "value": "AWS_US_WEST_2"},
        {"name": "snowflakeCommonConfig.snowflakeConnection.organization", "value": None},
        {"name": "emrServerlessConnection.awsConfig.awsSTSRegionConfig.signingRegion", "value": "US_WEST_2"},
        {"name": "emrServerlessConnection.awsConfig.awsSTSRegionConfig.regionDefinitionType", "value": "NOT_SPECIFIED"},
        {"name": "constants", "value": []},
        {"name": "snowflakeCommonConfig.snowflakeConnection.customSnowflakeRole", "value": None},
        {"name": "emrServerlessConnection.emrVersion", "value": "emr-6.8.0"},
        {"name": "clusterConfig.dataCacheLevel", "value": "MEMORY_AND_DISK"},
        {"name": "emrServerlessConnection.applicationId", "value": None},
        {"name": "databricksConfig.terminateCluster", "value": False},
        {"name": "snowflakeCommonConfig.snowflakeConnection.warehouse", "value": None},
        {"name": "transformerEmrConnection.bootstrapActions", "value": []},
        {"name": "transformerEmrConnection.primaryInstanceTypeCustom", "value": None},
        {"name": "googleCloudConfig.clusterLabels", "value": []},
        {"name": "snowflakeCommonConfig.snowflakeConnection.privateKeyPemPath", "value": None},
        {"name": "emrServerlessConnection.awsConfig.isAssumeRole", "value": False},
        {"name": "transformerEmrConnection.sshEc2KeyId", "value": None},
        {
            "name": "preprocessScript",
            "value": "/*\nThe following script define a method"
            "\nthat increments an integer by 1 "
            "\nand registers it as a UDF with "
            "\nthe SparkSession, which can be accessed"
            '\nusing the variable named "spark":'
            "\ndef inc(i: Integer): Integer = {\n  i + 1\n}"
            '\nspark.udf.register ("inc", inc _)\n\n*/',
        },
        {"name": "transformerEmrConnection.awsConfig.awsSTSRegionConfig.customRegion", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.useCustomUrl", "value": False},
        {"name": "databricksConfig.baseUrl", "value": None},
        {
            "name": "startEventStage",
            "value": "streamsets-datacollector-basic-lib::"
            "com_streamsets_pipeline_stage_destination_devnull_ToErrorNullDTarget::1",
        },
        {"name": "transformerEmrConnection.parameters", "value": []},
        {"name": "webhookConfigs", "value": []},
        {"name": "snowflakeCommonConfig.snowflakeConnection.customSnowflakeRegion", "value": None},
        {"name": "transformerEmrConnection.awsConfig.profileName", "value": None},
        {"name": "transformerEMRConfig.kmsKeyId", "value": None},
        {"name": "transformerEmrConnection.emrVersion", "value": None},
        {"name": "clusterConfig.callbackUrl", "value": None},
        {"name": "transformerEmrConnection.awsConfig.roleSessionName", "value": None},
        {"name": "clusterConfig.sparkAppName", "value": "${pipeline:title()}"},
        {
            "name": "transformerEmrConnection.awsConfig.awsSTSRegionConfig.regionDefinitionType",
            "value": "NOT_SPECIFIED",
        },
        {"name": "emrServerlessConnection.retryPolicyConfig.maxErrorRetry", "value": 6},
        {"name": "transformerEmrConnection.retryPolicyConfig.baseDelay", "value": 100},
        {"name": "emrServerlessConnection.awsConfig.awsSTSRegionConfig.region", "value": "US_WEST_2"},
        {"name": "clusterConfig.clusterType", "value": "LOCAL"},
        {"name": "clusterConfig.yarnKerberosPrincipal", "value": "name@DOMAIN"},
        {"name": "snowflakeCommonConfig.snowflakeConnection.oauthToken", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.useSnowflakeOrganizationURL", "value": True},
        {"name": "emrServerlessConnection.awsConfig.awsAccessKeyId", "value": None},
        {"name": "googleCloudConfig.terminate", "value": None},
        {"name": "cdeConfig.authTokenUrl", "value": None},
        {"name": "sparkConfigs", "value": []},
        {"name": "emrServerlessConnection.maximumDisk", "value": 20000},
        {"name": "transformerEmrConnection.versionName", "value": None},
        {"name": "emrServerlessConnection.runtimeRoleArn", "value": "arn:aws:iam::<account-id>:role/role-name"},
        {"name": "snowflakeCommonConfig.snowflakeConnection.connectionProperties", "value": None},
        {"name": "emrServerlessConnection.customRegion", "value": None},
        {"name": "emrServerlessConnection.retryPolicyConfig.maxBackoff", "value": 20000},
        {"name": "emrServerlessConnection.awsConfig.setSessionTags", "value": True},
        {"name": "emrServerlessConnection.awsConfig.sessionDuration", "value": 3600},
        {"name": "googleCloudConfig.clusterPrefix", "value": None},
        {"name": "snowflakeCommonConfig.overrideDb", "value": None},
        {"name": "emrServerlessConnection.awsConfig.awsSecretAccessKey", "value": None},
        {"name": "snowflakeCommonConfig.snowflakeConnection.oauthRedirectUri", "value": None},
        {"name": "transformerEmrConnection.awsConfig.roleARN", "value": "arn:aws:iam::<account-id>:role/role-name"},
        {"name": "emrServerlessConnection.securityGroupIds", "value": []},
    ],
    "uiInfo": {},
    "fragments": [],
    "stages": [],
    "errorStage": {
        "instanceName": "errorStageStageInstance",
        "library": "streamsets-datacollector-basic-lib",
        "stageName": "com_streamsets_pipeline_stage_destination_devnull_ToErrorNullDTarget",
        "stageVersion": "1",
        "configuration": [],
        "uiInfo": {"stageType": "TARGET", "label": "Error -Discard"},
        "inputLanes": [],
        "outputLanes": [],
        "eventLanes": [],
        "services": [],
    },
    "info": {
        "pipelineId": "use a flow's executor id here",
        "title": "use a flow's executor id here",
        "description": "description",  # ignore since it's only in the datacollector
        "created": 1758061153670,
        "lastModified": 1758061153670,
        "creator": "use the current user's iam_id here",
        "lastModifier": "use the current user's iam_id here",
        "lastRev": "0",
        "uuid": "create a new uuid here",
        "valid": False,
        "metadata": None,
        "name": "use a flow's executor id here",
        "sdcVersion": "7.0.0-SNAPSHOT",
        "sdcId": "dc74ef9f-0da1-4ee1-ac7b-177e95601edc",
    },
    "metadata": None,
    "statsAggregatorStage": {
        "instanceName": "statsAggregatorStageInstance",
        "library": "streamsets-datacollector-basic-lib",
        "stageName": "com_streamsets_pipeline_stage_destination_devnull_StatsDpmDirectlyDTarget",
        "stageVersion": "1",
        "configuration": [],
        "uiInfo": {
            "stageType": "TARGET",
            "label": "Stats Aggregator -Write Directly to Control Hub - "
            "statistics are not aggregated across Data Collectors",
        },
        "inputLanes": [],
        "outputLanes": [],
        "eventLanes": [],
        "services": [],
    },
    "startEventStages": [],
    "stopEventStages": [],
    "testOriginStage": {
        "instanceName": "com_streamsets_pipeline_stage_devtest_rawdata_RawDataDSource_TestOriginStage",
        "library": "streamsets-datacollector-dev-lib",
        "stageName": "com_streamsets_pipeline_stage_devtest_rawdata_RawDataDSource",
        "stageVersion": "4",
        "configuration": [
            {"name": "eventData", "value": None},
            {"name": "numberOfThreads", "value": 1},
            {"name": "rawData", "value": '{\n  "f1": "abc",\n  "f2": "xyz",\n  "f3": "lmn"\n}'},
            {"name": "stopAfterFirstBatch", "value": False},
            {"name": "stageOnRecordError", "value": "TO_ERROR"},
        ],
        "uiInfo": {"stageType": "SOURCE", "label": "Test Origin - Dev Raw Data Source"},
        "inputLanes": [],
        "outputLanes": ["com_streamsets_pipeline_stage_devtest_rawdata_RawDataDSource_TestOriginStageOutputLane1"],
        "eventLanes": [],
        "services": [
            {
                "service": "com.streamsets.pipeline.api.service.dataformats.DataFormatParserService",
                "serviceVersion": 4,
                "configuration": [
                    {"name": "displayFormats", "value": "DELIMITED,JSON,LOG,SDC_JSON,TEXT,XML"},
                    {"name": "dataFormat", "value": "JSON"},
                    {"name": "dataFormatConfig.compression", "value": "NONE"},
                    {"name": "dataFormatConfig.compressionLibrary", "value": "AUTODETECT"},
                    {"name": "dataFormatConfig.filePatternInArchive", "value": "*"},
                    {"name": "dataFormatConfig.charset", "value": "UTF-8"},
                    {"name": "dataFormatConfig.removeCtrlChars", "value": False},
                    {"name": "dataFormatConfig.textMaxLineLen", "value": 1024},
                    {"name": "dataFormatConfig.useCustomDelimiter", "value": False},
                    {"name": "dataFormatConfig.customDelimiter", "value": "\\r\\n"},
                    {"name": "dataFormatConfig.includeCustomDelimiterInTheText", "value": False},
                    {"name": "dataFormatConfig.jsonContent", "value": "MULTIPLE_OBJECTS"},
                    {"name": "dataFormatConfig.jsonMaxObjectLen", "value": 4096},
                    {"name": "dataFormatConfig.csvHeader", "value": "NO_HEADER"},
                    {"name": "dataFormatConfig.csvFileFormat", "value": "CSV"},
                    {"name": "dataFormatConfig.csvCustomDelimiter", "value": "|"},
                    {"name": "dataFormatConfig.multiCharacterFieldDelimiter", "value": "||"},
                    {"name": "dataFormatConfig.multiCharacterLineDelimiter", "value": "${str:unescapeJava('\\\\n')}"},
                    {"name": "dataFormatConfig.csvCustomEscape", "value": "\\"},
                    {"name": "dataFormatConfig.csvCustomQuote", "value": '"'},
                    {"name": "dataFormatConfig.csvUnivocityFieldSeparator", "value": ","},
                    {"name": "dataFormatConfig.csvUnivocityEscape", "value": "\\"},
                    {"name": "dataFormatConfig.csvUnivocityQuote", "value": '"'},
                    {"name": "dataFormatConfig.csvUnivocityLineSeparator", "value": "${str:unescapeJava('\\\\n')}"},
                    {"name": "dataFormatConfig.csvUnivocityAllowComments", "value": False},
                    {"name": "dataFormatConfig.csvUnivocityCommentMarker", "value": "#"},
                    {"name": "dataFormatConfig.csvEnableComments", "value": False},
                    {"name": "dataFormatConfig.csvCommentMarker", "value": "#"},
                    {"name": "dataFormatConfig.csvSkipStartLines", "value": 0},
                    {"name": "dataFormatConfig.csvParser", "value": "LEGACY_PARSER"},
                    {"name": "dataFormatConfig.csvAllowExtraColumns", "value": False},
                    {"name": "dataFormatConfig.csvExtraColumnPrefix", "value": "_extra_"},
                    {"name": "dataFormatConfig.csvMaxObjectLen", "value": 1024},
                    {"name": "dataFormatConfig.csvUnivocityMaxColumns", "value": 1000},
                    {"name": "dataFormatConfig.csvUnivocityMaxCharsPerColumn", "value": 1000},
                    {"name": "dataFormatConfig.csvUnivocitySkipEmptyLines", "value": False},
                    {"name": "dataFormatConfig.csvIgnoreEmptyLines", "value": False},
                    {"name": "dataFormatConfig.csvRecordType", "value": "LIST_MAP"},
                    {"name": "dataFormatConfig.parseNull", "value": False},
                    {"name": "dataFormatConfig.nullConstant", "value": "\\\\N"},
                    {"name": "dataFormatConfig.xmlRecordElement", "value": None},
                    {"name": "dataFormatConfig.preserveRootElement", "value": False},
                    {"name": "dataFormatConfig.includeFieldXpathAttributes", "value": False},
                    {"name": "dataFormatConfig.xPathNamespaceContext", "value": []},
                    {"name": "dataFormatConfig.outputFieldAttributes", "value": False},
                    {"name": "dataFormatConfig.xmlMaxObjectLen", "value": 4096},
                    {"name": "dataFormatConfig.logMode", "value": "COMMON_LOG_FORMAT"},
                    {"name": "dataFormatConfig.logMaxObjectLen", "value": 1024},
                    {"name": "dataFormatConfig.retainOriginalLine", "value": False},
                    {"name": "dataFormatConfig.customLogFormat", "value": '%h %l %u %t "%r" %>s %b'},
                    {
                        "name": "dataFormatConfig.regex",
                        "value": "^(\\S+) (\\S+) (\\S+) \\[([\\w:/]+\\s[+\\-]\\d{4})\\] "
                        '"(\\S+) (\\S+) (\\S+)" (\\d{3}) (\\d+)',
                    },
                    {"name": "dataFormatConfig.fieldPathsToGroupName", "value": [{"fieldPath": "/", "group": 1}]},
                    {"name": "dataFormatConfig.grokPatternDefinition", "value": None},
                    {"name": "dataFormatConfig.grokPattern", "value": "%{COMMONAPACHELOG}"},
                    {"name": "dataFormatConfig.onParseError", "value": "ERROR"},
                    {"name": "dataFormatConfig.maxStackTraceLines", "value": 50},
                    {"name": "dataFormatConfig.enableLog4jCustomLogFormat", "value": False},
                    {"name": "dataFormatConfig.log4jCustomLogFormat", "value": "%r [%t] %-5p %c %x - %m%n"},
                    {"name": "dataFormatConfig.avroSchemaSource", "value": None},
                    {"name": "dataFormatConfig.avroSchema", "value": None},
                    {"name": "dataFormatConfig.schemaRegistryUrls", "value": []},
                    {"name": "dataFormatConfig.schemaRegistrySecurityConfig.securityOption", "value": "PLAINTEXT"},
                    {"name": "dataFormatConfig.schemaRegistrySecurityConfig.truststoreType", "value": "JKS"},
                    {"name": "dataFormatConfig.schemaRegistrySecurityConfig.truststoreFile", "value": None},
                    {"name": "dataFormatConfig.schemaRegistrySecurityConfig.truststorePassword", "value": None},
                    {"name": "dataFormatConfig.schemaRegistrySecurityConfig.keystoreType", "value": "JKS"},
                    {"name": "dataFormatConfig.schemaRegistrySecurityConfig.keystoreFile", "value": None},
                    {"name": "dataFormatConfig.schemaRegistrySecurityConfig.keystorePassword", "value": None},
                    {"name": "dataFormatConfig.basicAuth", "value": None},
                    {"name": "dataFormatConfig.schemaLookupMode", "value": "SUBJECT"},
                    {"name": "dataFormatConfig.subject", "value": None},
                    {"name": "dataFormatConfig.schemaId", "value": None},
                    {"name": "dataFormatConfig.avroSkipUnionIndex", "value": False},
                    {"name": "dataFormatConfig.protoDescriptorFile", "value": None},
                    {"name": "dataFormatConfig.messageType", "value": None},
                    {"name": "dataFormatConfig.isDelimited", "value": False},
                    {"name": "dataFormatConfig.binaryMaxObjectLen", "value": 1024},
                    {"name": "dataFormatConfig.datagramMode", "value": "SYSLOG"},
                    {"name": "dataFormatConfig.typesDbPath", "value": None},
                    {"name": "dataFormatConfig.convertTime", "value": False},
                    {"name": "dataFormatConfig.excludeInterval", "value": False},
                    {"name": "dataFormatConfig.authFilePath", "value": None},
                    {"name": "dataFormatConfig.netflowOutputValuesMode", "value": "RAW_AND_INTERPRETED"},
                    {"name": "dataFormatConfig.maxTemplateCacheSize", "value": -1},
                    {"name": "dataFormatConfig.templateCacheTimeoutMs", "value": -1},
                    {"name": "dataFormatConfig.netflowOutputValuesModeDatagram", "value": "RAW_AND_INTERPRETED"},
                    {"name": "dataFormatConfig.maxTemplateCacheSizeDatagram", "value": -1},
                    {"name": "dataFormatConfig.templateCacheTimeoutMsDatagram", "value": -1},
                    {"name": "dataFormatConfig.wholeFileMaxObjectLen", "value": 8192},
                    {"name": "dataFormatConfig.rateLimit", "value": "-1"},
                    {"name": "dataFormatConfig.verifyChecksum", "value": False},
                    {"name": "dataFormatConfig.excelHeader", "value": None},
                    {"name": "dataFormatConfig.excelSkipCellsWithNoHeader", "value": False},
                    {"name": "dataFormatConfig.excelIncludeCellsWithEmptyValue", "value": False},
                    {"name": "dataFormatConfig.excelReadAllSheets", "value": False},
                    {"name": "dataFormatConfig.maximumFileSize", "value": 100},
                    {"name": "dataFormatConfig.excelSheetNames", "value": []},
                ],
            }
        ],
    },
    "valid": False,
    "issues": {
        "pipelineIssues": [
            {
                "instanceName": None,
                "serviceName": None,
                "level": "PIPELINE",
                "configGroup": None,
                "configName": None,
                "message": "VALIDATION_0001 - The pipeline is empty",
                "errorCode": "VALIDATION_0001",
                "humanReadableMessage": "The pipeline is empty",
                "technicalMessage": "The pipeline is empty",
                "remediationSteps": "",
                "count": 1,
                "additionalInfo": None,
            }
        ],
        "stageIssues": {},
        "issueCount": 1,
    },
    "previewable": False,
}


def create_pipeline_definition_for_new_flow(flow_executor_id: str, iam_id: str) -> dict:
    """Create an empty pipeline definition for a new flow, similar to what a data collector would.

    Args:
        flow_executor_id: A flow's id on a data collector
        iam_id: IAM ID of the user
    Returns:
        A `dict` which contains the pipeline definition for a new flow.
    """
    pipeline_uuid = str(uuid4())
    new_definition = deepcopy(EMPTY_PIPELINE_DEFINITION)

    # update the flow executor id
    new_definition["pipelineId"] = flow_executor_id
    new_definition["title"] = flow_executor_id
    new_definition["info"]["pipelineId"] = flow_executor_id
    new_definition["info"]["title"] = flow_executor_id
    new_definition["info"]["name"] = flow_executor_id

    # update the uuid
    new_definition["uuid"] = pipeline_uuid
    new_definition["info"]["uuid"] = pipeline_uuid

    # update iam id
    new_definition["info"]["creator"] = iam_id
    new_definition["info"]["lastModifier"] = iam_id

    return new_definition



# IBM Confidential
# PID 5900-BAF
# Copyright StreamSets Inc., an IBM Company 2025

"""This module contains API clients for apps common for all services across CPD."""

from ibm_watsonx_data_integration.cpd_api.access_groups_api import AccessGroupsAPIClient
from ibm_watsonx_data_integration.cpd_api.account_api import AccountAPIClient
from ibm_watsonx_data_integration.cpd_api.connections_api import ConnectionsApiClient
from ibm_watsonx_data_integration.cpd_api.global_catalog_api import GlobalCatalogApiClient
from ibm_watsonx_data_integration.cpd_api.global_search_api import GlobalSearchApiClient
from ibm_watsonx_data_integration.cpd_api.job_api import JobApiClient
from ibm_watsonx_data_integration.cpd_api.metering_api import MeteringApiClient
from ibm_watsonx_data_integration.cpd_api.project_api import ProjectApiClient
from ibm_watsonx_data_integration.cpd_api.resource_controller_api import ResourceControllerApiClient
from ibm_watsonx_data_integration.cpd_api.role_api import RoleAPIClient
from ibm_watsonx_data_integration.cpd_api.service_id_api import ServiceIDAPIClient
from ibm_watsonx_data_integration.cpd_api.trusted_profile_api import TrustedProfileAPIClient
from ibm_watsonx_data_integration.cpd_api.user_api import UserAPIClient

__all__ = [
    "AccountAPIClient",
    "GlobalCatalogApiClient",
    "GlobalSearchApiClient",
    "JobApiClient",
    "MeteringApiClient",
    "ProjectApiClient",
    "ProjectApiClient",
    "ResourceControllerApiClient",
    "UserAPIClient",
    "RoleAPIClient",
    "AccessGroupsAPIClient",
    "ConnectionsApiClient",
    "ServiceIDAPIClient",
    "TrustedProfileAPIClient",
]
